"""Test module for SLGOF export from GasOil"""

import hypothesis.strategies as st
import numpy as np
import pytest
from hypothesis import given

from pyscal import GasOil, WaterOilGas
from pyscal.constants import EPSILON, SWINTEGERS
from pyscal.utils.testing import sat_table_str_ok


def check_table(dframe):
    """Check sanity of important columns"""
    assert not dframe.empty
    assert not dframe.isna().to_numpy().any()
    assert dframe["SL"].is_monotonic_increasing
    assert (dframe["SL"] >= 0.0).all()
    assert (dframe["SL"] <= 1.0).all()
    # Increasing, but not monotonically for slgof
    assert (dframe["KROG"].diff().dropna() > -EPSILON).all()
    assert dframe["KRG"].is_monotonic_decreasing
    if "PC" in dframe:
        assert dframe["PC"].is_monotonic_decreasing


@given(
    st.floats(min_value=0.0, max_value=0.3),
    st.floats(min_value=0.0, max_value=0.3),
    st.floats(min_value=0.0, max_value=0.2),
)
def test_slgof(swl, sorg, sgcr):
    """Test dumping SLGOF records"""
    wog = WaterOilGas(swl=swl, sorg=sorg, sgcr=sgcr, h=0.05)
    wog.wateroil.add_corey_water()
    wog.wateroil.add_corey_oil()
    wog.gasoil.add_corey_gas(krgmax=1)
    wog.gasoil.add_corey_oil()

    assert wog.selfcheck()

    slgof = wog.gasoil.slgof_df()
    assert "SL" in slgof
    assert "KRG" in slgof
    assert "KROG" in slgof
    assert not slgof.empty

    check_table(slgof)
    sat_table_str_ok(wog.SLGOF())

    # Requirements from E100 manual:
    assert np.isclose(slgof["SL"].to_numpy()[0], wog.gasoil.swl + wog.gasoil.sorg)
    assert np.isclose(slgof["KRG"].to_numpy()[-1], 0)
    assert np.isclose(slgof["KROG"].to_numpy()[0], 0)

    # If we ruin the object, SLGOF() will return an empty string:
    wog.gasoil.table = wog.gasoil.table.drop("KRG", axis="columns")
    assert wog.gasoil.SLGOF() == ""
    assert wog.SLGOF() == ""


@pytest.mark.parametrize(
    "swl, sorg, sgcr",
    [
        (0.0, 0.0, 0.1),
        (1e-18, 0.0, 0.1),
        (1e-17, 0.0, 0.1),
        (1e-16, 0.0, 0.1),
        (1e-15, 0.0, 0.1),
        (1e-14, 0.0, 0.1),
        (1e-13, 0.0, 0.1),
        (1e-12, 0.0, 0.1),
        (1e-11, 0.0, 0.1),
        (1e-10, 0.0, 0.1),
        (0.1, 0.0, 0.1),
        (0.1, 1e-18, 0.1),
        (0.1, 1e-17, 0.1),
        (0.1, 1e-16, 0.1),
        (0.1, 1e-15, 0.1),
        (0.1, 1e-14, 0.1),
        (0.1, 1e-13, 0.1),
        (0.1, 1e-12, 0.1),
        (0.1, 1e-11, 0.1),
        (0.1, 1e-10, 0.1),
        (0.1, 0.0, 0.0),
        (0.1, 0.0, 1e-18),
        (0.1, 0.0, 1e-17),
        (0.1, 0.0, 1e-16),
        (0.1, 0.0, 1e-15),
        (0.1, 0.0, 1e-14),
        (0.1, 0.0, 1e-13),
        (0.1, 0.0, 1e-12),
        (0.1, 0.0, 1e-11),
        (0.1, 0.0, 1e-10),
        (0.1 - 0.0, 0.0, 0.1),
        (0.1 - 1e-18, 0.0, 0.1),
        (0.1 - 1e-17, 0.0, 0.1),
        (0.1 - 1e-16, 0.0, 0.1),
        (0.1 - 1e-15, 0.0, 0.1),
        (0.1 - 1e-14, 0.0, 0.1),
        (0.1 - 1e-13, 0.0, 0.1),
        (0.1 - 1e-12, 0.0, 0.1),
        (0.1 - 1e-11, 0.0, 0.1),
        (0.1 - 1e-10, 0.0, 0.1),
    ],
)
def test_numerical_problems(swl, sorg, sgcr):
    """Test fine-tuned numerics for slgof"""

    # Because we cut away some saturation points due to SWINTEGERS, we easily
    # end in a situation where the wrong saturation point of to "equal" ones
    # is removed (because in SLGOF, sg is flipped to sl)

    gasoil = GasOil(swl=swl, sorg=sorg, sgcr=sgcr, h=0.001)
    gasoil.add_corey_gas()
    gasoil.add_corey_oil()
    assert gasoil.selfcheck()
    slgof = gasoil.slgof_df()
    assert np.isclose(slgof["SL"].to_numpy()[0], gasoil.swl + gasoil.sorg)
    assert np.isclose(slgof["SL"].to_numpy()[-1], 1.0)
    check_table(slgof)


@given(
    st.floats(min_value=0.0, max_value=0.3),
    st.floats(min_value=0.0, max_value=0.3),
    st.floats(min_value=0.0, max_value=0.3),
    st.floats(min_value=1.0 / SWINTEGERS, max_value=0.5),
)
def test_slgof_hypo(swl, sorg, sgcr, h):
    """Shotgun-testing of slgof"""
    gasoil = GasOil(swl=swl, sorg=sorg, sgcr=sgcr, h=h)
    gasoil.add_corey_gas()
    gasoil.add_corey_oil()
    assert gasoil.selfcheck()
    slgof = gasoil.slgof_df()
    check_table(slgof)
    # Eclipse 100 requirement from manual:
    assert np.isclose(slgof["SL"].to_numpy()[0], gasoil.swl + gasoil.sorg)
    # Eclipse 100 requirement from manual:
    assert np.isclose(slgof["SL"].to_numpy()[-1], 1.0)
    slgof_str = gasoil.SLGOF()
    assert isinstance(slgof_str, str)
    assert slgof_str
    sat_table_str_ok(slgof_str)


@given(
    st.floats(min_value=0.0, max_value=0.3),
    st.floats(min_value=0.0, max_value=0.3),
    st.floats(min_value=1.0 / float(100 * SWINTEGERS), max_value=0.5),
)
def test_slgof_sl_mismatch(swl, sorg, h):
    """Test numerical robustness on slgof table creation."""
    gasoil = GasOil(h=h, swl=swl, sorg=sorg)
    gasoil.add_corey_gas()
    gasoil.add_corey_oil()

    # It is a strict requirement that the first sl value should be swl + sorg,
    # but GasOil might have truncated the values.
    assert np.isclose(
        gasoil.slgof_df()["SL"].to_numpy()[0] - (gasoil.swl + gasoil.sorg), 0
    )



"""Test module for the WaterOilGas object"""

import os
from pathlib import Path

import pytest

from pyscal import GasOil, WaterOil, WaterOilGas
from pyscal.utils.testing import sat_table_str_ok

try:
    import opm.io

    HAVE_OPM = True
except ImportError:
    HAVE_OPM = False


def test_wateroilgas_constructor():
    """Test that constructor properties are available as aattributes"""
    wog = WaterOilGas(swirr=0.01, swl=0.02, sorg=0.03, sorw=0.04, tag="Foo")
    assert wog.swirr == 0.01
    assert wog.swl == 0.02
    assert wog.sorg == 0.03
    assert wog.tag == "Foo"
    assert wog.sorw == 0.04

    # Manipulate the tag in the underlying GasOil object:
    wog.gasoil.tag = "Bar"
    assert wog.tag == "Foo Bar"  # Different tags are concatenated.


def test_wateroilgas_simple():
    """Test that default curves will give valid include strings"""
    wog = WaterOilGas()

    # Add default curves:
    wog.wateroil.add_corey_water()
    wog.wateroil.add_corey_oil()
    wog.gasoil.add_corey_gas()
    wog.gasoil.add_corey_oil()

    with pytest.raises(AssertionError):
        # Testing test code:
        sat_table_str_ok("")
    sat_table_str_ok(wog.SWOF())
    sat_table_str_ok(wog.SGOF())
    sat_table_str_ok(wog.SLGOF())
    sat_table_str_ok(wog.SOF3())
    sat_table_str_ok(wog.SGFN())
    sat_table_str_ok(wog.SWFN())


def test_threephasecheck():
    """Test three phase consistency checks"""
    wog = WaterOilGas()
    assert not wog.selfcheck()
    wog.wateroil.add_corey_water(nw=2)
    wog.wateroil.add_corey_oil(now=2, kroend=0.9)
    wog.gasoil.add_corey_gas(ng=2)
    wog.gasoil.add_corey_oil(nog=2, kroend=1)
    assert not wog.threephaseconsistency()


def test_empty():
    """Empty object should give empty strings (and logged errors)"""
    wog = WaterOilGas()
    assert wog.SWOF() == ""
    assert wog.SGOF() == ""
    assert wog.SOF3() == ""
    assert wog.SLGOF() == ""
    assert wog.SWFN() == ""
    assert wog.SGFN() == ""


def test_manipulated_attributes_none():
    """It is allowed to manipulate the WaterOilGas by setting the wateroil
    and/or the gasoil attributes to None, this is performed in
    SCALrecommendation.interpolate() for example. Test this behaviour."""

    wog_go = WaterOilGas(tag="gasoilonly")
    # Make it into a two-phase object:
    wog_go.wateroil = None
    wog_go.gasoil.add_corey_gas()
    wog_go.gasoil.add_corey_oil()
    assert wog_go.SWFN() == ""
    assert wog_go.SWOF() == ""
    assert wog_go.SOF3() == ""
    with pytest.raises(ValueError, match="wateroil is None in WaterOilGas"):
        wog_go.swirr
    with pytest.raises(ValueError, match="wateroil is None in WaterOilGas"):
        wog_go.swl
    with pytest.raises(ValueError, match="wateroil is None in WaterOilGas"):
        wog_go.sorw
    assert wog_go.tag == "gasoilonly"
    assert wog_go.threephaseconsistency() is True

    wog_wo = WaterOilGas(tag="wateroilonly")
    wog_wo.gasoil = None
    wog_wo.wateroil.add_corey_water()
    wog_wo.wateroil.add_corey_oil()
    assert wog_wo.SLGOF() == ""
    assert wog_wo.SGFN() == ""
    assert wog_wo.SOF3() == ""
    assert wog_wo.SGOF() == ""
    with pytest.raises(ValueError, match="gasoil is None in WaterOilGas"):
        wog_wo.sorg
    assert wog_wo.tag == "wateroilonly"
    assert wog_wo.threephaseconsistency() is True

    wog_nones = WaterOilGas()
    wog_nones.wateroil = None
    wog_nones.gasoil = None
    wog_nones.tag == ""
    assert wog_nones.threephaseconsistency() is True
    assert wog_nones.SWOF() == ""
    assert wog_nones.SGOF() == ""
    assert wog_nones.selfcheck() is False
    with pytest.raises(ValueError, match="wateroil is None"):
        wog_nones.swirr


def test_not_threephase_consistency():
    """Mock a WaterOilGas object that fails threephase consistency"""
    wog = WaterOilGas()
    # To trigger this, we need to hack the WaterOilGas object
    # by overriding the effect of its __init__
    wog.wateroil = WaterOil(swl=0.4)
    wog.gasoil = GasOil(swl=0.2)
    wog.wateroil.add_corey_water(nw=2)
    wog.wateroil.add_corey_oil(now=2, kroend=0.9)
    wog.gasoil.add_corey_gas(ng=2)
    wog.gasoil.add_corey_oil(nog=2, kroend=1)
    assert not wog.threephaseconsistency()


@pytest.mark.skipif(not HAVE_OPM, reason="ecl2df not installed")
def test_parse_with_opm(tmp_path):
    """Test that the SWOF+SGOF output from pyscal can be
    injected into a valid Eclipse deck"""
    wog = WaterOilGas()
    wog.wateroil.add_corey_water(nw=2)
    wog.wateroil.add_corey_oil(now=2, kroend=0.9)
    wog.gasoil.add_corey_gas(ng=2)
    wog.gasoil.add_corey_oil(nog=2, kroend=1)

    ecldeck = (
        """RUNSPEC
DIMENS
  1 1 1 /
OIL
WATER
GAS
START
  1 'JAN' 2100 /
TABDIMS
   2* 10000 /
EQLDIMS
  1 /
GRID
DX
   10 /
DY
   10 /
DZ
   50 /
TOPS
   1000 /
PORO
   0.3 /
PERMX
   100 /
PERMY
   100 /
PERMZ
   100 /

PROPS

"""
        + wog.SWOF()
        + wog.SGOF()
        + """
DENSITY
  800 1000 1.2 /
PVTW
  1 1 0.0001 0.2 0.00001 /
PVDO
   100 1   1
   150 0.9 1 /
PVDG
   100 1 1
   150 0.9 1 /
ROCK
  100 0.0001 /
SOLUTION
EQUIL
   1000    100     1040    0   1010      0 /"""
    )

    os.chdir(tmp_path)
    Path("RELPERMTEST.DATA").write_text(ecldeck, encoding="utf8")
    deck = opm.io.Parser().parse("RELPERMTEST.DATA")
    assert "SWOF" in deck
    assert "SGOF" in deck



"""Generates plots for usage in documentation"""

from pathlib import Path

import matplotlib.pyplot as plt

from pyscal import GasOil, GasWater, WaterOil

IMG_DIR = Path(__file__).absolute().parent / "images"


def main():
    """Run this to produce the PNG images for the RST docs"""

    for plotname in [
        # Three first plots are there to show all possible
        # object parameterizations:
        "gasoil-endpoints",
        "gaswater-endpoints",
        "wateroil-endpoints",
        # Remaining plots are for dedicated flowing processes:
        "wateroil-idc2",
        "gasoil-cdi2",
        "gaswater-icd2",
        "gasoil-cid2",
        "gaswater-dci1",
        "gaswater-co2-icd2",
        "gaswater-paleogas-dci3",
        "wateroil-paleooil-idc1",
    ]:
        assert "_" not in plotname, "Keep names consistent please.."
        eval("make_" + plotname.replace("-", "_") + "(show=False)")
        plt.tight_layout()
        plt.savefig(IMG_DIR / (plotname + ".png"), dpi=200)


def make_gasoil_endpoints(show=True, krgendanchor="sorg"):
    """Make a plot explaining the inputs to a GasOil object"""
    plt.xkcd()
    _, axes = plt.subplots()
    swl = 0.1
    sgcr = 0.1
    sgro = 0.1
    sorg = 0.2
    krgend = 0.7
    krgmax = 0.75
    kroend = 0.8
    kromax = 0.85
    gasoil = GasOil(sgcr=sgcr, sorg=sorg, swl=swl, sgro=sgro, krgendanchor=krgendanchor)
    gasoil2 = GasOil(sgcr=sgcr, sorg=sorg, swl=swl, sgro=sgro, krgendanchor=None)
    gasoil.add_corey_gas(ng=2, krgend=krgend, krgmax=krgmax)
    gasoil2.add_corey_gas(ng=2, krgend=krgend)
    gasoil.add_corey_oil(nog=2, kroend=kroend, kromax=kromax)
    gasoil2.add_corey_oil(nog=2, kroend=kroend, kromax=kromax)
    gasoil2.table.plot(
        ax=axes, x="SG", y="KRG", c="pink", alpha=0.7, label="KRG*", linewidth=2
    )
    gasoil.table.plot(
        ax=axes, x="SG", y="KRG", c="red", alpha=1, label="KRG", linewidth=2
    )
    gasoil.table.plot(
        ax=axes, x="SG", y="KROG", c="green", alpha=1, label="KROG", linewidth=2
    )
    plt.ylim([-0.02, 1])
    plt.xticks([0, 1])
    plt.yticks([0, 1])
    axes.annotate(
        "KROMAX",
        xy=(0, kromax),
        arrowprops={"arrowstyle": "->"},
        xytext=(0.1, kromax + 0.07),
    )
    axes.annotate(
        "KROEND",
        xy=(sgro, kroend),
        arrowprops={"arrowstyle": "->"},
        xytext=(sgro + 0.07, kroend + 0.05),
    )
    axes.annotate(
        "KRGEND",
        xy=(1 - sorg - swl, krgend),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - sorg - swl + 0.01, krgmax - 0.2),
    )
    # Overwrite with same text at identical spots for
    # two arrows:
    axes.annotate(
        "KRGEND",
        xy=(1 - swl, krgend),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - sorg - swl + 0.01, krgmax - 0.2),
    )
    axes.annotate(
        "KRGMAX",
        xy=(1 - swl, krgmax),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - swl - 0.2, krgmax + 0.1),
    )
    axes.text(0.0, 0.07, "SGCR")
    axes.annotate(
        "", xy=(-0.01, 0.04), xytext=(sgcr, 0.04), arrowprops={"arrowstyle": "<->"}
    )
    axes.text(0.0, 0.73, "SGRO")
    axes.annotate(
        "",
        xy=(-0.01, kroend - 0.02),
        xytext=(sgro, kroend - 0.02),
        arrowprops={"arrowstyle": "<->"},
    )
    plt.xlabel("SG", labelpad=-10)
    axes.annotate(
        "",
        xy=(1 - sorg - swl, 0.02),
        xytext=(1 - swl, 0.02),
        arrowprops={"arrowstyle": "<->"},
    )
    axes.text(1 - sorg - swl + 0.04, 0.04, "SORG")
    axes.annotate(
        "", xy=(1 - swl, 0.02), xytext=(1, 0.02), arrowprops={"arrowstyle": "<->"}
    )
    axes.text(1 - swl + 0.04, 0.04, "SWL")
    axes.legend(loc="upper center")
    if show:
        plt.show()


def make_wateroil_endpoints(show=True):
    """Make a plot explaining the inputs to a WaterOil object"""
    plt.xkcd()
    _, axes = plt.subplots()
    swirr = 0.05
    swl = 0.1
    swcr = 0.2
    sorw = 0.2
    krwend = 0.7
    krwmax = 0.75
    kroend = 0.85
    wateroil = WaterOil(swirr=swirr, swl=swl, swcr=swcr, sorw=sorw)
    wateroil.add_corey_water(nw=2, krwend=krwend, krwmax=krwmax)
    wateroil.add_corey_oil(now=2, kroend=kroend)
    wateroil.table.plot(
        ax=axes, x="SW", y="KRW", c="blue", alpha=1, label="KRW", linewidth=2
    )
    wateroil.table.plot(
        ax=axes, x="SW", y="KROW", c="green", alpha=1, label="KROW", linewidth=2
    )
    plt.ylim([-0.02, 1])
    plt.xticks([0, 1])
    plt.yticks([0, 1])
    axes.annotate(
        "KROEND",
        xy=(swl, kroend),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl - 0.1, kroend - 0.3),
    )
    axes.annotate(
        "KRWEND",
        xy=(1 - sorw, krwend),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - sorw - 0.09, krwmax - 0.2),
    )
    axes.annotate(
        "KRWMAX",
        xy=(1, krwmax),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - 0.1, krwmax - 0.3),
    )
    axes.annotate(
        "SWIRR",
        xy=(swirr, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swirr - 0.12, 0 + 0.1),
    )
    axes.annotate(
        "SWL",
        xy=(swl, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl - 0.05, 0 + 0.14),
    )
    axes.annotate(
        "SWCR",
        xy=(swcr, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swcr - 0.06, 0 + 0.21),
    )
    plt.xlabel("SW", labelpad=-10)
    axes.annotate(
        "", xy=(1 - sorw, 0.02), xytext=(1, 0.02), arrowprops={"arrowstyle": "<->"}
    )
    axes.text(1 - sorw + 0.04, 0.04, "SORW")
    if show:
        plt.show()


def make_gaswater_endpoints(show=True):
    """Make a plot explaining the inputs to a WaterOil object"""
    plt.xkcd()
    _, axes = plt.subplots()
    swirr = 0.05
    swl = 0.1
    swcr = 0.2
    sgcr = 0.4
    sgrw = 0.2
    krwend = 0.65
    krwmax = 0.75
    krgend = 0.85
    gaswater = GasWater(swirr=swirr, swl=swl, swcr=swcr, sgrw=sgrw, sgcr=sgcr, h=0.001)
    gaswater.add_corey_water(nw=2, krwend=krwend, krwmax=krwmax)
    gaswater.add_corey_gas(ng=2, krgend=krgend)
    gaswater.wateroil.table.plot(
        ax=axes, x="SW", y="KRW", c="blue", alpha=1, label="KRW", linewidth=2
    )
    gaswater.gasoil.table.plot(
        ax=axes, x="SL", y="KRG", c="red", alpha=1, label="KRG", linewidth=2
    )
    plt.ylim([-0.02, 1])
    plt.xticks([0, 1])
    plt.yticks([0, 1])
    axes.annotate(
        "KRGEND",
        xy=(swl, krgend),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl - 0.1, krgend - 0.3),
    )
    axes.annotate(
        "KRWEND",
        xy=(1 - sgrw, krwend),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - sgrw - 0.23, krwmax - 0.2),
    )
    axes.annotate(
        "KRWMAX",
        xy=(1, krwmax),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - 0.1, krwmax - 0.3),
    )
    axes.annotate(
        "SWIRR",
        xy=(swirr, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swirr - 0.12, 0 + 0.1),
    )
    axes.annotate(
        "SWL",
        xy=(swl, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl - 0.05, 0 + 0.14),
    )
    axes.annotate(
        "SWCR",
        xy=(swcr, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swcr - 0.06, 0 + 0.21),
    )
    plt.xlabel("SW", labelpad=-10)

    axes.text(1 - sgrw + 0.04, 0.04, "SGRW")
    axes.annotate(
        "", xy=(1 - sgrw, 0.02), xytext=(1, 0.02), arrowprops={"arrowstyle": "<->"}
    )

    axes.text(1 - sgcr + 0.04, 0.14, "SGCR")
    axes.annotate(
        "", xy=(1 - sgcr, 0.12), xytext=(1, 0.12), arrowprops={"arrowstyle": "<->"}
    )

    if show:
        plt.show()


def make_wateroil_idc2(show=True):
    plt.xkcd()
    _, axes = plt.subplots()
    swirr = 0.05
    swl = 0.1
    swcr = 0.1
    sorw = 0.25
    krwend = 0.8
    krwmax = 1
    kroend = 0.85
    wateroil = WaterOil(swirr=swirr, swl=swl, swcr=swcr, sorw=sorw)
    wateroil.add_corey_water(nw=2, krwend=krwend, krwmax=krwmax)
    wateroil.add_corey_oil(now=1.8, kroend=kroend)
    wateroil.table.plot(
        ax=axes, x="SW", y="KRW", c="blue", alpha=1, label="KRW", linewidth=2
    )
    wateroil.table.plot(
        ax=axes, x="SW", y="KROW", c="green", alpha=1, label="KROW", linewidth=2
    )
    plt.ylim([-0.02, 1.02])
    plt.xlim([-0.02, 1.02])
    plt.xticks([0, 1])
    plt.yticks([0, 1])
    axes.annotate(
        "KROEND",
        xy=(swl, kroend),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl + 0.1, kroend + 0.1),
    )
    axes.annotate(
        "KRWEND",
        xy=(1 - sorw, krwend),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - 0.15, krwend - 0.02),
    )
    axes.annotate(
        "KRWMAX",
        xy=(1, krwmax),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - 0.3, krwmax - 0.02),
    )
    axes.annotate(
        "SWIRR",
        xy=(swirr, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swirr - 0.12, 0 + 0.1),
    )
    axes.annotate(
        "SWL≈SWCR",
        xy=(swl, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl - 0.055, 0 + 0.14),
    )
    plt.xlabel("SW", labelpad=-10)
    axes.legend(loc="upper center")
    axes.annotate(
        "", xy=(1 - sorw, 0.02), xytext=(1, 0.02), arrowprops={"arrowstyle": "<->"}
    )
    axes.text(1 - sorw + 0.04, 0.04, "SORW")

    # Initial state and saturation direction:
    plt.vlines(swl, ymin=0, ymax=1, colors="darkorange", linestyles="dashed")
    axes.arrow(
        swl + 0.03,
        kroend - 0.01,
        0.05,
        -0.09,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )
    axes.arrow(
        swl + 0.03,
        0.03,
        0.05,
        0.01,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )

    if show:
        plt.show()


def make_gasoil_cdi2(show=True):
    plt.xkcd()
    _, axes = plt.subplots()
    swl = 0.1
    sgcr = 0.2
    sorg = 0.2
    krgend = 0.8
    krgmax = 1
    kroend = 0.85
    gasoil = GasOil(sgcr=sgcr, sorg=sorg, swl=swl, krgendanchor="sorg")
    gasoil.add_corey_gas(ng=3, krgend=krgend, krgmax=krgmax)
    gasoil.add_corey_oil(nog=3, kroend=kroend)
    gasoil.table.plot(
        ax=axes, x="SG", y="KRG", c="red", alpha=1, label="KRG", linewidth=2
    )
    gasoil.table.plot(
        ax=axes, x="SG", y="KROG", c="green", alpha=1, label="KROG", linewidth=2
    )
    plt.ylim([-0.02, 1])
    plt.xlim([-0.02, 1])
    plt.xticks([0, 1])
    plt.yticks([0, 1])
    axes.annotate(
        "KROEND",
        xy=(0, kroend),
        arrowprops={"arrowstyle": "->"},
        xytext=(0.1, kroend + 0.02),
    )
    axes.annotate(
        "KRGEND",
        xy=(1 - sorg - swl, krgend),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - 0.2, krgend - 0.02),
    )
    axes.annotate(
        "KRGMAX",
        xy=(1 - swl, krgmax),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - 0.35, krgmax - 0.04),
    )
    axes.annotate(
        "SGCR",
        xy=(sgcr, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(sgcr - 0.05, 0 + 0.14),
    )
    plt.xlabel("SG", labelpad=-10)
    axes.annotate(
        "",
        xy=(1 - sorg - swl, 0.02),
        xytext=(1 - swl, 0.02),
        arrowprops={"arrowstyle": "<->"},
    )
    axes.text(1 - sorg - swl + 0.04, 0.04, "SORG")
    axes.annotate(
        "", xy=(1 - swl, 0.02), xytext=(1, 0.02), arrowprops={"arrowstyle": "<->"}
    )
    axes.text(1 - swl + 0.04, 0.04, "SWL")
    axes.legend(loc="upper center")

    # Initial state and saturation direction:
    plt.vlines(0.0, ymin=0, ymax=1, colors="darkorange", linestyles="dashed")
    axes.arrow(
        0.03,
        kroend - 0.03,
        0.08,
        -0.2,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )
    axes.arrow(
        0.03,
        0.03,
        0.1,
        0.001,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )
    if show:
        plt.show()


def make_gasoil_cid2(show=True):
    plt.xkcd()
    _, axes = plt.subplots()
    swl = 0.1
    sgcr = 0.15
    sgro = 0.15
    sorg = 0.15
    krgend = 1
    krgmax = 1
    kroend = 0.8
    kromax = 0.9
    gasoil = GasOil(sgcr=sgcr, sorg=sorg, sgro=sgro, swl=swl, krgendanchor="")
    gasoil.add_corey_gas(ng=3, krgend=krgend, krgmax=krgmax)
    gasoil.add_corey_oil(nog=3, kroend=kroend, kromax=kromax)
    gasoil.table.plot(
        ax=axes, x="SG", y="KRG", c="red", alpha=1, label="KRG", linewidth=2
    )
    gasoil.table.plot(
        ax=axes, x="SG", y="KROG", c="green", alpha=1, label="KROG", linewidth=2
    )
    plt.ylim([-0.02, 1])
    plt.xlim([-0.02, 1])
    plt.xticks([0, 1])
    plt.yticks([0, 1])
    axes.annotate(
        "KROEND",
        xy=(sgro, kroend),
        arrowprops={"arrowstyle": "->"},
        xytext=(sgro + 0.1, kroend - 0.05),
    )
    axes.annotate(
        "KROMAX",
        xy=(0, kromax),
        arrowprops={"arrowstyle": "->"},
        xytext=(0.1, kromax + 0.05),
    )
    axes.annotate(
        "KRGEND",
        xy=(1 - swl, krgend - 0.01),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - 0.35, krgend - 0.04),
    )
    axes.annotate(
        "SGCR=SGRO",
        xy=(sgcr, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(sgcr - 0.1, 0 + 0.14),
    )
    plt.xlabel("SG", labelpad=-10)
    axes.annotate(
        "",
        xy=(1 - sorg - swl, 0.08),
        xytext=(1 - swl, 0.08),
        arrowprops={"arrowstyle": "<->"},
    )
    axes.text(1 - sorg - swl + 0.02, 0.1, "SORG")
    axes.annotate(
        "", xy=(1 - swl, 0.02), xytext=(1, 0.02), arrowprops={"arrowstyle": "<->"}
    )
    axes.text(1 - swl + 0.04, 0.04, "SWL")
    axes.legend(loc="upper center")

    # Initial state and saturation direction:
    plt.vlines(1 - swl, ymin=0, ymax=1, colors="darkorange", linestyles="dashed")
    axes.arrow(
        1 - swl - 0.03,
        krgend - 0.03,
        -0.08,
        -0.2,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )
    axes.arrow(
        1 - swl - 0.03,
        0.03,
        -0.13,
        0.001,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )
    if show:
        plt.show()


def make_gaswater_icd2(show=True):
    plt.xkcd()
    _, axes = plt.subplots()
    swl = 0.1
    swcr = 0.2
    sgcr = 0.2
    sgrw = 0.2
    krwend = 0.65
    krwmax = 0.75
    krgend = 1
    gaswater = GasWater(swl=swl, swcr=swcr, sgrw=sgrw, sgcr=sgcr, h=0.001)
    gaswater.add_corey_water(nw=2, krwend=krwend, krwmax=krwmax)
    gaswater.add_corey_gas(ng=2, krgend=krgend)
    gaswater.wateroil.table.plot(
        ax=axes, x="SW", y="KRW", c="blue", alpha=1, label="KRW", linewidth=2
    )
    gaswater.gasoil.table.plot(
        ax=axes, x="SL", y="KRG", c="red", alpha=1, label="KRG", linewidth=2
    )
    plt.ylim([-0.02, 1])
    plt.xticks([0, 1])
    plt.yticks([0, 1])
    axes.annotate(
        "KRGEND",
        xy=(swl, krgend),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl + 0.1, krgend - 0.1),
    )
    axes.annotate(
        "KRWEND",
        xy=(1 - sgrw, krwend),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - sgrw - 0.23, krwmax - 0.1),
    )
    axes.annotate(
        "KRWMAX",
        xy=(1 - 0.01, krwmax),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - 0.23, krwmax),
    )
    axes.annotate(
        "SWL",
        xy=(swl, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl - 0.07, 0 + 0.12),
    )
    axes.annotate(
        "SWCR",
        xy=(swcr, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swcr - 0.06, 0 + 0.21),
    )
    plt.xlabel("SW", labelpad=-10)

    axes.text(1 - sgrw + 0.04, 0.04, "SGRW")
    axes.annotate(
        "", xy=(1 - sgrw, 0.02), xytext=(1, 0.02), arrowprops={"arrowstyle": "<->"}
    )

    axes.text(1 - sgcr + 0.04, 0.14, "SGCR")
    axes.annotate(
        "", xy=(1 - sgcr, 0.12), xytext=(1, 0.12), arrowprops={"arrowstyle": "<->"}
    )

    # Initial state and saturation direction:
    plt.vlines(swl, ymin=0, ymax=1, colors="darkorange", linestyles="dashed")
    axes.arrow(
        swl + 0.03,
        krgend - 0.03,
        0.08,
        -0.2,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )
    axes.arrow(
        swl + 0.03,
        0.03,
        0.1,
        0.005,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )

    if show:
        plt.show()


def make_gaswater_dci1(show=True):
    # CO2 storage
    plt.xkcd()
    _, axes = plt.subplots()
    swl = 0.1
    swcr = 0.1
    sgcr = 0.2
    sgrw = 0.0
    krwend = 1
    krgend = 0.8
    gaswater = GasWater(swl=swl, swcr=swcr, sgrw=sgrw, sgcr=sgcr, h=0.001)
    gaswater.add_corey_water(nw=2, krwend=krwend)
    gaswater.add_corey_gas(ng=2, krgend=krgend)
    gaswater.wateroil.table.plot(
        ax=axes, x="SW", y="KRW", c="blue", alpha=1, label="KRW", linewidth=2
    )
    gaswater.gasoil.table.plot(
        ax=axes, x="SL", y="KRG", c="red", alpha=1, label="KRG", linewidth=2
    )
    plt.ylim([-0.02, 1])
    plt.xticks([0, 1])
    plt.yticks([0, 1])
    plt.xlim([-0.02, 1.02])
    axes.annotate(
        "KRGEND",
        xy=(swl, krgend),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl + 0.1, krgend - 0.1),
    )
    axes.annotate(
        "KRWEND",
        xy=(1 - sgrw, krwend),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - sgrw - 0.23, krwend - 0.1),
    )
    axes.annotate(
        "SWL=SWCR",
        xy=(swl, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl - 0.08, 0 + 0.1),
    )
    plt.xlabel("SW", labelpad=-10)

    # axes.text(1 - sgrw + 0.04, 0.04, "SGRW")
    axes.annotate(
        "", xy=(1 - sgrw, 0.02), xytext=(1, 0.02), arrowprops={"arrowstyle": "<->"}
    )

    axes.text(1 - sgcr + 0.04, 0.14, "SGCR")
    axes.annotate(
        "", xy=(1 - sgcr, 0.12), xytext=(1, 0.12), arrowprops={"arrowstyle": "<->"}
    )

    # Initial state and saturation direction:
    plt.vlines(1, ymin=0, ymax=1, colors="darkorange", linestyles="dashed")
    axes.arrow(
        1 - 0.03,
        krwend - 0.03,
        -0.08,
        -0.14,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )
    axes.arrow(
        1 - 0.03,
        0.03,
        -0.1,
        0.005,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )
    axes.legend(loc="upper center")

    if show:
        plt.show()


def make_gaswater_co2_icd2(show=True):
    # CO2 trapping
    plt.xkcd()
    _, axes = plt.subplots()
    swl = 0.1
    swcr = 0.1
    sgcr = 0.3
    sgrw = 0.3
    krwend = 0.3
    krwmax = 1
    krgend = 0.8
    gaswater = GasWater(swl=swl, swcr=swcr, sgrw=sgrw, sgcr=sgcr, h=0.001)
    gaswater.add_corey_water(nw=3, krwend=krwend, krwmax=krwmax)
    gaswater.add_corey_gas(ng=2, krgend=krgend)
    gaswater.wateroil.table.plot(
        ax=axes, x="SW", y="KRW", c="blue", alpha=1, label="KRW", linewidth=2
    )
    gaswater.gasoil.table.plot(
        ax=axes, x="SL", y="KRG", c="red", alpha=1, label="KRG", linewidth=2
    )
    plt.ylim([-0.02, 1])
    plt.xticks([0, 1])
    plt.yticks([0, 1])
    plt.xlim([-0.02, 1.02])
    axes.annotate(
        "KRGEND",
        xy=(swl, krgend),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl + 0.1, krgend - 0.1),
    )
    axes.annotate(
        "KRWEND",
        xy=(1 - sgrw, krwend),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - sgrw - 0.23, krwend + 0.1),
    )
    axes.annotate(
        "KRWMAX",
        xy=(1, krwmax),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - 0.23, krwmax - 0.1),
    )
    axes.annotate(
        "SWL=SWCR",
        xy=(swl, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl - 0.1, 0 + 0.1),
    )
    plt.xlabel("SW", labelpad=-10)

    axes.text(1 - sgrw + 0.04, 0.04, "SGRW=SGCR")
    axes.annotate(
        "", xy=(1 - sgrw, 0.02), xytext=(1, 0.02), arrowprops={"arrowstyle": "<->"}
    )

    # Initial state and saturation direction:
    sw_initial = 0.3
    plt.vlines(sw_initial, ymin=0, ymax=1, colors="darkorange", linestyles="dashed")
    axes.arrow(
        sw_initial + 0.01,
        0.37,
        0.08,
        -0.12,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )
    axes.arrow(
        sw_initial + 0.01,
        0.03,
        0.1,
        0.03,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )
    axes.legend(loc="upper center")

    if show:
        plt.show()


def make_wateroil_paleooil_idc1(show=True):
    plt.xkcd()
    _, axes = plt.subplots()
    swl = 0.1
    swcr = 0.2
    sorw = 0.15
    socr = 0.25
    krwend = 0.8
    krwmax = 1
    kroend = 0.85
    wateroil = WaterOil(swl=swl, swcr=swcr, sorw=sorw, socr=socr)
    wateroil.add_corey_water(nw=2, krwend=krwend, krwmax=krwmax)
    wateroil.add_corey_oil(now=2, kroend=kroend)
    wateroil.table.plot(
        ax=axes, x="SW", y="KRW", c="blue", alpha=1, label="KRW", linewidth=2
    )
    wateroil.table.plot(
        ax=axes, x="SW", y="KROW", c="green", alpha=1, label="KROW", linewidth=2
    )
    plt.ylim([-0.02, 1])
    plt.xticks([0, 1])
    plt.yticks([0, 1])
    axes.annotate(
        "KROEND",
        xy=(swl, kroend),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl + 0.07, kroend),
    )
    axes.annotate(
        "KRWEND",
        xy=(1 - sorw, krwend),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - sorw - 0.15, krwend + 0.05),
    )
    axes.annotate(
        "KRWMAX",
        xy=(1, krwmax),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - 0.25, krwmax - 0.05),
    )
    axes.annotate(
        "SWL",
        xy=(swl, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl - 0.05, 0 + 0.14),
    )
    axes.annotate(
        "SWCR",
        xy=(swcr, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swcr - 0.06, 0 + 0.21),
    )
    plt.xlabel("SW", labelpad=-10)
    axes.annotate(
        "", xy=(1 - sorw, krwend), xytext=(1, krwend), arrowprops={"arrowstyle": "<->"}
    )
    axes.text(1 - sorw + 0.03, krwend - 0.05, "SORW")
    axes.annotate(
        "",
        xy=(1 - socr, 0.12),
        xytext=(1, 0.12),
        arrowprops={"arrowstyle": "<->"},
    )
    axes.text(1 - socr + 0.05, 0.14, "SOCR")
    axes.legend(loc="upper center")

    # Initial state and saturation direction:
    plt.vlines(1 - sorw, ymin=0, ymax=1, colors="darkorange", linestyles="dashed")
    axes.arrow(
        1 - sorw - 0.03,
        krwend - 0.03,
        -0.05,
        -0.09,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )
    axes.arrow(
        1 - sorw - 0.03,
        0.03,
        -0.05,
        0.01,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )

    if show:
        plt.show()


def make_gaswater_paleogas_dci3(show=True):
    plt.xkcd()
    _, axes = plt.subplots()
    swl = 0.1
    sgl = 0.1
    swcr = 0.2
    sgcr = 0.3
    sgrw = sgl
    krwend = 0.85
    krgend = 0.9
    gaswater = GasWater(swl=swl, sgl=sgl, swcr=swcr, sgrw=sgrw, sgcr=sgcr, h=0.001)
    gaswater.add_corey_water(nw=3, krwend=krwend)
    gaswater.add_corey_gas(ng=2, krgend=krgend)
    gaswater.wateroil.table.plot(
        ax=axes, x="SW", y="KRW", c="blue", alpha=1, label="KRW", linewidth=2
    )
    gaswater.gasoil.table.plot(
        ax=axes, x="SL", y="KRG", c="red", alpha=1, label="KRG", linewidth=2
    )
    plt.ylim([-0.02, 1])
    plt.xlim([0, 1.02])
    plt.xticks([0, 1])
    plt.yticks([0, 1])
    axes.annotate(
        "KRGEND",
        xy=(swl, krgend),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl - 0.1, krgend - 0.3),
    )
    axes.annotate(
        "KRWEND",
        xy=(1 - sgrw, krwend),
        arrowprops={"arrowstyle": "->"},
        xytext=(1 - sgrw - 0.23, krwend - 0.1),
    )
    axes.annotate(
        "SWL=SWCR",
        xy=(swl, 0),
        arrowprops={"arrowstyle": "->"},
        xytext=(swl - 0.1, 0 + 0.14),
    )
    plt.xlabel("SW", labelpad=-10)
    axes.legend(loc="upper center")

    axes.text(1 - sgrw + 0.01, krwend - 0.05, "SGRW")
    axes.annotate(
        "", xy=(1 - sgrw, krwend), xytext=(1, krwend), arrowprops={"arrowstyle": "<->"}
    )

    axes.text(1 - sgcr + 0.04, 0.09, "SGCR")
    axes.annotate(
        "", xy=(1 - sgcr, 0.07), xytext=(1, 0.07), arrowprops={"arrowstyle": "<->"}
    )

    axes.text(1 - sgl + 0.02, 0.02, "SGL")
    axes.annotate(
        "", xy=(1 - sgl, 0.00), xytext=(1, 0.00), arrowprops={"arrowstyle": "<->"}
    )

    # Initial state and saturation direction:
    plt.vlines(1 - sgl, ymin=0, ymax=1, colors="darkorange", linestyles="dashed")
    # Upper arrow, along water curve:
    axes.arrow(
        1 - sgl - 0.03,
        krwend - 0.06,
        -0.05,
        -0.13,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )
    # Lower arrow, along gas curve:
    axes.arrow(
        1 - sgl - 0.03,
        0.03,
        -0.08,
        0.00,
        head_width=0.025,
        fill=True,
        facecolor="darkorange",
    )
    if show:
        plt.show()


if __name__ == "__main__":
    main()



"""Test module for SCAL recommendation objects"""

import hypothesis.strategies as st
import pandas as pd
import pytest
from hypothesis import given

from pyscal import (
    GasWater,
    SCALrecommendation,
    WaterOil,
    WaterOilGas,
)
from pyscal.factory import (
    create_scal_recommendation,
    create_scal_recommendation_list,
    create_water_oil_gas,
    load_relperm_df,
    slicedict,
)
from pyscal.utils.testing import check_table, sat_table_str_ok, slow_hypothesis

# Example SCAL recommendation, low case
LOW_SAMPLE_LET = {
    "swirr": 0.1,
    "sorw": 0.02,
    "krwend": 0.7,
    "krwmax": 0.8,
    "swl": 0.16,
    "swcr": 0.25,
    "Lw": 2.323,
    "Ew": 2.0,
    "Tw": 1.329,
    "Lo": 4.944,
    "Eo": 5.0,
    "To": 0.68,
    "Lg": 4,
    "Eg": 1,
    "Tg": 1,
    "Log": 4,
    "Eog": 1,
    "Tog": 1,
    "sorg": 0.2,
    "sgcr": 0.15,
    "krgend": 0.9,
    "krgmax": 1,
    "kroend": 1,
    "tag": "SATNUM X",
}
# Example SCAL recommendation, base case
BASE_SAMPLE_LET = {
    "swirr": 0.1,
    "sorw": 0.091,
    "krwend": 0.8,
    "swl": 0.16,
    "swcr": 0.20,
    "Lw": 3.369,
    "Ew": 4.053,
    "Tw": 1.047,
    "Lo": 3.726,
    "Eo": 3.165,
    "To": 1.117,
    "Lg": 2,
    "Eg": 2,
    "Tg": 2,
    "Log": 2,
    "Eog": 2,
    "Tog": 2,
    "sorg": 0.1,
    "sgcr": 0.10,
    "krgend": 0.97,
    "kroend": 1,
    "tag": "SATNUM X",
}
# Example SCAL recommendation, high case
HIGH_SAMPLE_LET = {
    "swirr": 0.1,
    "sorw": 0.137,
    "krwend": 0.6,
    "swl": 0.16,
    "swcr": 0.16,
    "Lw": 4.436,
    "Ew": 8.0,
    "Tw": 0.766,
    "Lo": 2.537,
    "Eo": 2.0,
    "To": 1.549,
    "Lg": 1,
    "Eg": 2,
    "Tg": 2,
    "Log": 1,
    "Eog": 2,
    "Tog": 2,
    "sorg": 0.05,
    "sgcr": 0.0,
    "krgend": 1,
    "kroend": 1,
    "tag": "SATNUM X",
}


def test_make_scalrecommendation():
    """Test that we can make scal recommendation objects
    from three WaterOilGas objects"""

    low = create_water_oil_gas(LOW_SAMPLE_LET)
    base = create_water_oil_gas(BASE_SAMPLE_LET)
    high = create_water_oil_gas(HIGH_SAMPLE_LET)
    rec = SCALrecommendation(low, base, high)
    interpolant = rec.interpolate(-0.5, h=0.2)
    check_table(interpolant.wateroil.table)
    check_table(interpolant.gasoil.table)

    # Check preservation of tags.
    swof = interpolant.SWOF()
    assert "SCAL recommendation interpolation to -0.5" in swof
    assert "SCAL recommendation interpolation to -0.5" in interpolant.SGOF()
    assert "SCAL recommendation interpolation to -0.5" in interpolant.SWFN()
    assert "SCAL recommendation interpolation to -0.5" in interpolant.SOF3()
    assert "SCAL recommendation interpolation to -0.5" in interpolant.SGFN()

    assert swof.count("SATNUM X") == 1
    assert interpolant.SOF3().count("SATNUM X") == 1

    # Check different comment when different interpolation parameter:
    interpolant = rec.interpolate(-0.777, 0.888, h=0.2)
    assert "SCAL recommendation interpolation to -0.777" in interpolant.SWOF()
    assert "SCAL recommendation interpolation to 0.888" in interpolant.SGOF()

    with pytest.raises(ValueError, match="Wrong arguments to SCALrecommendation"):
        SCALrecommendation([], [], [])


def test_make_scalrecommendation_wo(caplog):
    """Test that we can make scal recommendation objects
    from three WaterOilGas objects, but only with WaterOil
    objects"""

    wo_param_names = [
        "swirr",
        "sorw",
        "krwend",
        "krwmax",
        "swl",
        "swcr",
        "Lw",
        "Ew",
        "Tw",
        "Lo",
        "Eo",
        "To",
        "kroend",
    ]

    low_let_wo = slicedict(LOW_SAMPLE_LET, wo_param_names)
    low = create_water_oil_gas(low_let_wo)
    base_let_wo = slicedict(BASE_SAMPLE_LET, wo_param_names)
    base = create_water_oil_gas(base_let_wo)
    high_let_wo = slicedict(HIGH_SAMPLE_LET, wo_param_names)
    assert "Lg" not in high_let_wo
    high = create_water_oil_gas(high_let_wo)
    rec = SCALrecommendation(low, base, high)
    interpolant = rec.interpolate(-0.5)
    check_table(interpolant.wateroil.table)
    assert interpolant.gasoil is None
    sat_table_str_ok(interpolant.SWOF())
    sat_table_str_ok(interpolant.SWFN())

    # This should return empty string
    assert not interpolant.SGOF()

    # Meaningless extra argument is ignored, only warning printed:
    rec.interpolate(-1, parameter2=1)
    assert "parameter2 is meaningless for water-oil only" in caplog.text


def test_make_scalrecommendation_go():
    """Test that we can make scal recommendation objects
    from three WaterOilGas objects, but only with GasOil
    objects"""

    go_param_names = [
        "swirr",
        "sorg",
        "krgend",
        "krgmax",
        "swl",
        "sgcr",
        "Lg",
        "Eg",
        "Tg",
        "Log",
        "Eog",
        "Tog",
        "kroend",
    ]

    low_let_go = slicedict(LOW_SAMPLE_LET, go_param_names)
    low = create_water_oil_gas(low_let_go)
    base_let_go = slicedict(BASE_SAMPLE_LET, go_param_names)
    base = create_water_oil_gas(base_let_go)
    high_let_go = slicedict(HIGH_SAMPLE_LET, go_param_names)
    assert "Lw" not in high_let_go
    high = create_water_oil_gas(high_let_go)
    rec = SCALrecommendation(low, base, high)
    assert rec.type == WaterOilGas
    interpolant = rec.interpolate(-0.5)
    check_table(interpolant.gasoil.table)
    assert interpolant.wateroil is None
    sat_table_str_ok(interpolant.SGOF())
    sat_table_str_ok(interpolant.SGFN())

    # This should return empty string
    assert not interpolant.SWOF()

    with pytest.raises(ValueError, match="Interpolation parameter for gas must be in"):
        rec.interpolate(2)


@slow_hypothesis
@given(
    st.floats(min_value=-1.1, max_value=1.1), st.floats(min_value=-1.1, max_value=1.1)
)
def test_interpolation(param_wo, param_go):
    """Test interpolation with random interpolation parameters,
    looking for numerical corner cases"""

    rec = create_scal_recommendation(
        {"low": LOW_SAMPLE_LET, "base": BASE_SAMPLE_LET, "high": HIGH_SAMPLE_LET},
        "foo",
        h=0.1,
    )
    rec.add_simple_J()  # Add default pc curve

    # Check that added pc curve is non-zero
    assert sum(rec.low.wateroil.table["PC"])
    assert sum(rec.base.wateroil.table["PC"])
    assert sum(rec.high.wateroil.table["PC"])

    try:
        interpolant = rec.interpolate(param_wo, param_go, h=0.1)
    except ValueError:
        return

    check_table(interpolant.wateroil.table)
    check_table(interpolant.gasoil.table)

    assert len(interpolant.gasoil.SGOF()) > 100
    assert len(interpolant.gasoil.SGFN()) > 100
    assert len(interpolant.wateroil.SWFN()) > 100
    assert len(interpolant.SOF3()) > 100
    assert len(interpolant.wateroil.SWOF()) > 100
    assert interpolant.threephaseconsistency()

    assert sum(interpolant.wateroil.table["PC"])


def test_boundary_cases():
    """Test that interpolation is able to return the boundaries
    at +/- 1"""
    rec = create_scal_recommendation(
        {"low": LOW_SAMPLE_LET, "base": BASE_SAMPLE_LET, "high": HIGH_SAMPLE_LET},
        "foo",
        h=0.1,
    )
    assert rec.type == WaterOilGas

    wo_cols = ["SW", "KRW", "KROW"]  # no Pc in this test data
    go_cols = ["SG", "KRG", "KROG"]
    assert (
        rec.interpolate(0)
        .wateroil.table[wo_cols]
        .equals(rec.base.wateroil.table[wo_cols])
    )
    assert (
        rec.interpolate(-1)
        .wateroil.table[wo_cols]
        .equals(rec.low.wateroil.table[wo_cols])
    )
    assert (
        rec.interpolate(1)
        .wateroil.table[wo_cols]
        .equals(rec.high.wateroil.table[wo_cols])
    )

    assert (
        rec.interpolate(0).gasoil.table[go_cols].equals(rec.base.gasoil.table[go_cols])
    )
    assert (
        rec.interpolate(-1).gasoil.table[go_cols].equals(rec.low.gasoil.table[go_cols])
    )
    assert (
        rec.interpolate(1).gasoil.table[go_cols].equals(rec.high.gasoil.table[go_cols])
    )

    assert (
        rec.interpolate(0, 1)
        .wateroil.table[wo_cols]
        .equals(rec.base.wateroil.table[wo_cols])
    )
    assert (
        rec.interpolate(-1, 1)
        .wateroil.table[wo_cols]
        .equals(rec.low.wateroil.table[wo_cols])
    )
    assert (
        rec.interpolate(1, 1)
        .wateroil.table[wo_cols]
        .equals(rec.high.wateroil.table[wo_cols])
    )

    assert (
        rec.interpolate(0, 1)
        .gasoil.table[go_cols]
        .equals(rec.high.gasoil.table[go_cols])
    )
    assert (
        rec.interpolate(-1, 1)
        .gasoil.table[go_cols]
        .equals(rec.high.gasoil.table[go_cols])
    )
    assert (
        rec.interpolate(1, 1)
        .gasoil.table[go_cols]
        .equals(rec.high.gasoil.table[go_cols])
    )

    assert (
        rec.interpolate(0, 0)
        .gasoil.table[go_cols]
        .equals(rec.base.gasoil.table[go_cols])
    )
    assert (
        rec.interpolate(-1, 0)
        .gasoil.table[go_cols]
        .equals(rec.base.gasoil.table[go_cols])
    )
    assert (
        rec.interpolate(1, 0)
        .gasoil.table[go_cols]
        .equals(rec.base.gasoil.table[go_cols])
    )

    assert (
        rec.interpolate(0, -1)
        .gasoil.table[go_cols]
        .equals(rec.low.gasoil.table[go_cols])
    )
    assert (
        rec.interpolate(-1, -1)
        .gasoil.table[go_cols]
        .equals(rec.low.gasoil.table[go_cols])
    )
    assert (
        rec.interpolate(1, -1)
        .gasoil.table[go_cols]
        .equals(rec.low.gasoil.table[go_cols])
    )


def test_corner_errors():
    """Test more or less likely error scenarios"""

    # This could possibly also have been exception scenarios:
    assert (
        SCALrecommendation(GasWater(), GasWater(), GasWater()).interpolate(0).SWFN()
        == ""
    )
    assert (
        SCALrecommendation(WaterOilGas(), WaterOilGas(), WaterOilGas())
        .interpolate(0)
        .SWFN()
        == ""
    )

    with pytest.raises(ValueError, match="Wrong arguments to SCALrecommendation"):
        SCALrecommendation(WaterOil(), WaterOil(), WaterOil())

    with pytest.raises(ValueError, match="Wrong arguments to SCALrecommendation"):
        SCALrecommendation(WaterOilGas(), WaterOilGas(), GasWater())


def test_gaswater_scal(caplog):
    """Test list of gas-water objects in scal recommendation"""
    dframe = pd.DataFrame(
        columns=["SATNUM", "CASE", "NW", "NG", "TAG"],
        data=[
            [1, "pess", 2, 2, "sometag"],
            [1, "base", 3, 3, "sometag"],
            [1, "opt", 4, 4, "sometag"],
        ],
    )
    rec_list = create_scal_recommendation_list(load_relperm_df(dframe), h=0.1)
    assert rec_list.pyscaltype == SCALrecommendation
    assert rec_list[1].type == GasWater
    low_list = rec_list.interpolate(-1)
    str_fam2 = low_list.build_eclipse_data(family=2)
    assert "SCAL recommendation interpolation to -1" in str_fam2
    assert "SGFN" in str_fam2
    assert "SWFN" in str_fam2

    # Meaningless extra argument is ignored, only warning printed:
    rec_list[1].interpolate(-1, parameter2=1)
    assert "parameter2 is meaningless for gas-water" in caplog.text
    caplog.clear()

    rec_list.interpolate(-1, int_params_go=1)
    assert "parameter2 is meaningless for gas-water" in caplog.text

    with pytest.raises(ValueError, match="Interpolation parameter must be in"):
        rec_list[1].interpolate(2)


def test_fast():
    """Test the fast option"""
    low_fast = create_water_oil_gas(LOW_SAMPLE_LET, fast=True)
    base_fast = create_water_oil_gas(BASE_SAMPLE_LET, fast=True)
    high_fast = create_water_oil_gas(HIGH_SAMPLE_LET, fast=True)

    rec = SCALrecommendation(low_fast, base_fast, high_fast)
    interp = rec.interpolate(-0.5)
    assert rec.fast
    assert interp.fast

    # test that one or more inputs not being set to fast does not trigger fast mode
    low = create_water_oil_gas(LOW_SAMPLE_LET)
    base = create_water_oil_gas(BASE_SAMPLE_LET)
    high = create_water_oil_gas(HIGH_SAMPLE_LET)

    rec = SCALrecommendation(low_fast, base_fast, high)
    interp = rec.interpolate(-0.5)
    assert not rec.fast
    assert not interp.fast

    rec = SCALrecommendation(low, base_fast, high)
    interp = rec.interpolate(-0.5)
    assert not rec.fast
    assert not interp.fast

    rec = SCALrecommendation(low, base, high)
    interp = rec.interpolate(-0.5)
    assert not rec.fast
    assert not interp.fast



"""Monotonocity support functions for pyscal"""

import logging
import sys
from typing import Dict, List, Union

import numpy as np
import pandas as pd

from pyscal.constants import EPSILON as epsilon

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


logger = logging.getLogger(__name__)


class MonotonicitySpec(TypedDict, total=False):
    """Specification of monotonicity for a vector of values"""

    sign: int
    """Value of +1 dictates strictly increasing,
    value of -1 dictates scrictly decreasing. Required parameter."""

    upper: float
    """Values will be clipped at upper limit, and non-strict monotonicity is
    allowed at limit. Optional parameter."""

    lower: float
    """Values will be clipped at lower limit, and non-strict monotonicity is
    allowed at limit. Optional parameter."""

    allowzero: bool
    """If True, consecutive zeros will be allowes in an otherwise strictly
    monotonic column. Optional parameter."""


def modify_dframe_monotonicity(
    dframe: pd.DataFrame, monotonicity: Dict[str, MonotonicitySpec], digits: int
) -> pd.DataFrame:
    """Modify a dataframe for monotonicity.

    Columns in the dataframe are modified in-place.

    Number intervals to consider when enforcing monotonicity::

      <value>                          <orig>    <fixed>
      <lower limit>                     0.00      0.00
      <values smaller than accuracy>    0.0002    0.00
      <accuracy limit>                  0.01      0.01
      <potential constants>             0.010001  0.02
      <allow ups/downs below accuracy>  0.0100001 0.03
                                        0.01      0.04
      <upper limit minus accuracy>      0.99      0.99
      <values too close to upper limit> 0.999     1.00
      <overshooting values>             1.0001    1.00
      <upper limit>                     1.00      1.00

    Values close to  upper or lower limits (if limits are
    supplied), but which deviate from the limit by less
    than the requested accuracy are allowed, and will be
    shifted to the limits.

    Only strict monotonicity is supported. Non-strict
    monotonicity is only allowed at upper and lower limit, or
    for all-zero vectors if that option is activated.

    For non-strict monotocity, see the function clip_accumulate()

    Args:
        dframe: Data to modify.
        monotonicity: Keys are column names
        digits: Number of digits to ensure monotonicity for.
    """
    validate_monotonicity_arg(monotonicity, dframe.columns)

    # Wateroil.SWOF() (and similar) supply a column view
    # of the internal wateroil.table dataframe. When asked
    # to enforce monotonicity, it must be done on a copy, both
    # for speed and for not compromising the original data.

    # Round to an accuracy one notch finer than end results,
    # to avoid representation errors:
    dframe = dframe.round(digits + 1)

    # Prepare and check columns:
    for col, spec in monotonicity.items():
        if dframe[col].dtype != np.float64:
            dframe[col] = dframe[col].astype(float)

        # Bail on clearly erroneous data:
        check_almost_monotone(dframe[col], digits, spec["sign"])

        check_limits(dframe[col], spec)

    # Modify data for monotonicity:
    for col, spec in monotonicity.items():
        accuracy = 1.0 / 10.0**digits - epsilon

        if "allowzero" in spec:
            # Treat zero as an exception for strict monotonicity:
            max_value = dframe[col].abs().max()
            if max_value < accuracy and spec["allowzero"]:
                continue

        constants = rows_to_be_fixed(dframe[col], spec, digits)
        iterations = 0
        sign = spec["sign"]
        while constants.any():
            iterations += 1

            assert iterations <= 2 * len(dframe[col]), (
                "Too many iterations for monotonicity fix"
            )

            dframe.loc[constants, col] = (
                dframe.loc[constants, col] + sign / 10.0**digits - epsilon
            )

            # Ensure nonstrict monotonicity and clips after each modification:
            dframe[col] = clip_accumulate(dframe[col], spec)

            # Evaluate what is left to fix:
            constants = rows_to_be_fixed(dframe[col], spec, digits)

        # Warn if more iterations than 5% of the rows
        # (number of iterations do not necessarily correspond with
        # number of changed rows)
        if float(iterations) / float(len(dframe[col])) > 0.05:
            logger.warning(
                "Needed %s iterations on column %s of length %s",
                str(iterations),
                col,
                str(len(dframe[col])),
            )

        # Assert that we have successfully managed to force monotonicity
        allowance = 1.0 / 10.0**digits
        if sign > 0:
            assert not (dframe[col].round(digits).diff() < -allowance).any(), (
                "Not possible to make column monotonically increasing"
            )
        else:
            assert not (dframe[col].round(digits).diff() > allowance).any(), (
                "Not possible to make column monotonically decreasing"
            )
    return dframe


def clip_accumulate(
    series: Union[List[float], pd.Series, np.ndarray], monotonicity: MonotonicitySpec
) -> np.ndarray:
    """
    Modify a series (vector of numbers) for non-strict monotonicity, and
    optionally clip at lower and upper limits.

    Args:
        series: Vector of numbers to modify
        monotonicity:

    Returns:
        np.array, copy of original.
    """
    series = np.array(series)
    if monotonicity["sign"] > 0:
        series = np.maximum.accumulate(series)
    else:
        series = np.minimum.accumulate(series)
    if "lower" in monotonicity and "upper" in monotonicity:
        series = series.clip(min=monotonicity["lower"], max=monotonicity["upper"])
    elif "lower" in monotonicity:
        series = series.clip(min=monotonicity["lower"])
    elif "upper" in monotonicity:
        series = series.clip(max=monotonicity["upper"])
    return series


def check_limits(
    series: Union[List[float], pd.Series, np.ndarray],
    monotonicity: MonotonicitySpec,
    colname: str = "",
) -> None:
    """
    Check a series whether it obeys numerical limits.
    Equivalence to limits is allowed.

    Exceptions will be raised in case of error. Nothing is returned
    when everything is ok.

    Args:
        series: Vector of numbers to check
        monotonicity:
        colname: Optional string for a column name that will be
            included in any error message.
    """
    if isinstance(series, (list, np.ndarray)):
        series = pd.Series(series, dtype="float64")
    if series.empty:
        return
    if "upper" in monotonicity and (series > monotonicity["upper"]).any():
        raise ValueError(f"Values larger than upper limit in column {colname}")
    if "lower" in monotonicity and (series < monotonicity["lower"]).any():
        raise ValueError(f"Values smaller than lower limit in column {colname}")


def rows_to_be_fixed(
    series: pd.Series, monotonicity: MonotonicitySpec, digits: int
) -> pd.Series:
    """Compute boolean array of rows that must be modified

    Args:
        series:
        monotonicity:
        digits: Accuracy required, how many digits
            that are to be printed, and to which we should relate
            constancy to.
    Returns:
        boolean series.
    """
    if isinstance(series, (list, np.ndarray)):
        series = pd.Series(series, dtype="float64")

    # minus epsilon is critical to avoid being greedy
    accuracy = 1.0 / 10.0**digits - epsilon
    if monotonicity["sign"] > 0:
        constants = series.round(digits + 1).diff() < accuracy
    else:
        constants = series.round(digits + 1).diff() > -accuracy

    # Allow constants at the lower and upper limits.
    if "upper" in monotonicity:
        constants = constants & (series < (monotonicity["upper"] - accuracy))
    if "lower" in monotonicity:
        constants = constants & (series > (monotonicity["lower"] + accuracy))
    return constants


def check_almost_monotone(series: pd.Series, digits: int, sign: int) -> None:
    """Raise a ValueError if a series is not sufficiently close
    to constant or monotone in a certain direction.

    Args:
        series: Vector of numbers
        digits:
        sign: direction. >0 means positive
    """
    if isinstance(series, (list, np.ndarray)):
        series = pd.Series(series, dtype="float64")

    allowance = 1.0 / 10.0 ** (digits - 1)
    if sign > 0:
        if series.diff().min() < -allowance:
            raise ValueError("Series is not almost monotone")
    elif series.diff().max() > allowance:
        raise ValueError("Series is not almost monotone")


def validate_monotonicity_arg(
    monotonicity: Dict[str, MonotonicitySpec], dframe_colnames: List[str]
) -> None:
    """
    Validate a dictionary with monotonicity arguments that
    can be given to df2str().

    Will raise ValueError exceptions if anything is wrong.

    Args:
        monotonicity: Keys are column names.
        dframe_colnames: Names of column names
            in dframes. Used in error messages.
    """
    valid_keys = ["sign", "upper", "lower", "allowzero"]
    if monotonicity is None:
        return
    if not isinstance(monotonicity, dict):
        raise ValueError("monotonicity argument must be a dict")
    for col, spec in monotonicity.items():
        if not isinstance(spec, dict):
            raise ValueError("monotonicity argument must be a dict of dicts")
        if not set(spec.keys()).issubset(valid_keys):
            raise ValueError(f"Unknown keys in monotonicity {spec.keys()}")
        if col not in dframe_colnames:
            raise ValueError(f"Column {col} does not exist in dataframe")
        if "sign" not in spec:
            raise ValueError(f"Monotonocity sign not specified for {col}")
        try:
            signvalue = float(spec["sign"])
        except ValueError as err:
            raise ValueError(f"Monotonocity sign {spec['sign']} not valid") from err
        if "upper" in spec:
            float(spec["upper"])
        if "lower" in spec:
            float(spec["lower"])
        if abs(signvalue) > 1:
            raise ValueError("Monotonocity sign must be -1 or +1, not larger/smaller")

        if "allowzero" in spec and spec["allowzero"] not in {
            True,
            False,
        }:
            raise ValueError("allowzero in monotonicity argument must be True/False")



"""This file gets auto-discovered by pytest, and is the recommended
place to have common fixtures, and avoiding flake8 complaints"""

import logging

import pytest
from hypothesis import HealthCheck, settings

settings.register_profile(
    "ci", max_examples=250, deadline=None, suppress_health_check=[HealthCheck.too_slow]
)


@pytest.fixture
def default_loglevel():
    """Reset the log level for all registered loggers to WARNING

    This is necessary when testing for appearance of different log
    messages, when some of the tested commands might manipulate the log
    levels in order to set INFO or DEBUG messages, e.g. through --verbose
    sent to the command line client"""
    loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
    for logger in loggers:
        logger.setLevel(logging.WARNING)
    return default_loglevel


def pytest_addoption(parser):
    """Add option(s) that can be used on the pytest command line"""
    parser.addoption(
        "--plot",
        action="store_true",
        default=False,
        help="Run tests that display plots to the screen",
    )


def pytest_collection_modifyitems(config, items):
    """Act on options set on the command line"""
    if config.getoption("--plot"):
        # Do not skip tests when --plot is supplied on pytest command line
        return
    skip_plot = pytest.mark.skip(reason="need --plot option to run")
    for item in items:
        if "plot" in item.keywords:
            item.add_marker(skip_plot)



"""Constants used for pyscal modules:

* ``SWINTEGERS``: Number of different Sw values within [0,1] we allow
  This is used to create integer indices of Sw, since Floating Point
  indices are flaky in Pandas (and in general on computers)

* ``EPSILON``: Used as "a small number" for ensuring no floating point
  comparisons/errors pop up.  You cannot have the h parameter less
  than this when generating relperm tables

* ``MAX_EXPONENT_KR``: Maximal number for exponents in relperm parameterizations.
  Used to avoid numerical instabilities. It could probably be much
  higher than the chosen number in most circumstances, but such high
  numbers should not be relevant for relative permeability

* ``MAX_EXPONENT_PC``: Maximal number for exponents in cap pressure parameterizations.
"""

SWINTEGERS: int = 100000

EPSILON: float = 1e-08

MAX_EXPONENT_KR: float = 100

MAX_EXPONENT_PC: float = 10000



"""Test module for the WaterOil object"""

import io
import sys

import hypothesis.strategies as st
import matplotlib
import matplotlib.pyplot
import numpy as np
import pandas as pd
import pytest
from hypothesis import given

from pyscal import WaterOil
from pyscal.constants import SWINTEGERS
from pyscal.utils.testing import (
    check_linear_sections,
    check_table,
    float_df_checker,
    sat_table_str_ok,
)


def check_endpoints(wateroil, krwend, krwmax, kroend):
    """Check that the code produces correct endpoints for
    parametrizations, on discrete cases"""
    swtol = 1 / SWINTEGERS

    # Check endpoints for oil curve:
    assert float_df_checker(wateroil.table, "SW", wateroil.swl, "KROW", kroend)
    assert float_df_checker(wateroil.table, "SON", 1, "KROW", kroend)
    assert np.isclose(wateroil.table["KROW"].max(), kroend)
    assert float_df_checker(wateroil.table, "SON", 0.0, "KROW", 0.0)
    assert float_df_checker(wateroil.table, "SW", 1 - wateroil.socr, "KROW", 0.0)

    # Check endpoints for water curve: (np.isclose is only reliable around 1)
    assert float_df_checker(wateroil.table, "SWN", 0.0, "KRW", 0.0)
    assert float_df_checker(wateroil.table, "SW", wateroil.swcr, "KRW", 0)

    if wateroil.sorw > swtol:
        # (hard to get it right when sorw is less than h and close to zero)
        assert float_df_checker(wateroil.table, "SW", 1 - wateroil.sorw, "KRW", krwend)
        assert np.isclose(wateroil.table["KRW"].max(), krwmax)
    else:
        assert np.isclose(wateroil.table["KRW"].max(), krwend)


@given(st.text())
def test_wateroil_tag(tag):
    """Test that we are unlikely to crash Eclipse
    by having ugly tag names"""
    wateroil = WaterOil(h=0.5, tag=tag)
    wateroil.add_corey_oil()
    wateroil.add_corey_water()
    sat_table_str_ok(wateroil.SWOF())
    sat_table_str_ok(wateroil.SWFN())


@given(st.floats(), st.floats())
def test_wateroil_corey1(nw, now):
    """Test random corey parameters"""
    wateroil = WaterOil()
    try:
        wateroil.add_corey_oil(now=now)
        wateroil.add_corey_water(nw=nw)
    except AssertionError:
        # This happens for "invalid" input
        return

    assert "KROW" in wateroil.table
    assert "KRW" in wateroil.table
    assert isinstance(wateroil.krwcomment, str)
    check_table(wateroil.table)
    check_linear_sections(wateroil)
    swofstr = wateroil.SWOF()
    assert len(swofstr) > 100


@given(st.floats(), st.floats(), st.floats(), st.floats(), st.floats())
def test_wateroil_let1(l, e, t, krwend, krwmax):
    """Test random LET parameters"""
    wateroil = WaterOil()
    try:
        wateroil.add_LET_oil(l, e, t, krwend)
        wateroil.add_LET_water(l, e, t, krwend, krwmax)
    except AssertionError:
        # This happens for negative values f.ex.
        return
    assert "KROW" in wateroil.table
    assert "KRW" in wateroil.table
    assert isinstance(wateroil.krwcomment, str)
    check_table(wateroil.table)
    check_linear_sections(wateroil)
    swofstr = wateroil.SWOF()
    assert len(swofstr) > 100


@given(
    st.floats(min_value=0, max_value=0.4),
    st.floats(min_value=0, max_value=0.4),
    st.floats(min_value=0, max_value=0.4),
    st.floats(min_value=0, max_value=0.1),
    st.floats(min_value=0.1, max_value=1),
    st.floats(min_value=0.1, max_value=1),
    st.floats(min_value=0.1, max_value=1),
    st.floats(min_value=0.001, max_value=1),
    st.booleans(),
)
def test_wateroil_krendmax(swl, swcr, sorw, socr_add, kroend, krwend, krwmax, h, fast):
    """Test endpoints for wateroil using hypothesis testing"""
    try:
        wateroil = WaterOil(
            swl=swl, swcr=swcr, socr=sorw + socr_add, sorw=sorw, h=h, fast=fast
        )
    except AssertionError:
        return
    krwend = min(krwend, krwmax)
    wateroil.add_corey_oil(kroend=kroend)
    wateroil.add_corey_water(krwend=krwend, krwmax=krwmax)
    check_table(wateroil.table)
    assert wateroil.selfcheck()
    assert 0 < wateroil.crosspoint() < 1

    check_endpoints(wateroil, krwend, krwmax, kroend)
    ####################################
    # Do it over again, but with LET:
    wateroil.add_LET_oil(t=1.1, kroend=kroend)
    wateroil.add_LET_water(t=1.1, krwend=krwend, krwmax=krwmax)
    assert wateroil.selfcheck()
    check_table(wateroil.table)
    # Check endpoints for oil curve:
    check_endpoints(wateroil, krwend, krwmax, kroend)
    check_linear_sections(wateroil)
    assert 0 < wateroil.crosspoint() < 1


def test_fast():
    """Test the fast option"""
    # First without fast:
    wateroil = WaterOil(h=0.1)
    wateroil.add_corey_water()
    wateroil.add_corey_oil()
    # This crosspoint computation is only present for fast=False:
    assert "-- krw = krow @ sw=0.5" in wateroil.SWOF()

    # Provoke non-strict-monotone krow:
    wateroil.table.loc[0:2, "KROW"] = [1.00, 0.81, 0.81]
    # (this is valid in non-imbibition, but pyscal will correct it for all
    # curves)
    assert "0.1000000 0.0100000 0.8100000 0.0000000" in wateroil.SWOF()
    assert "0.2000000 0.0400000 0.8099999 0.0000000" in wateroil.SWOF()
    #   monotonicity correction:   ^^^^^^

    # Now redo with fast option:
    wateroil = WaterOil(h=0.1, fast=True)
    wateroil.add_corey_water()
    wateroil.add_corey_oil()
    # This crosspoint computation is only present for fast=False:
    assert "-- krw = krow" not in wateroil.SWOF()

    # Provoke non-strict-monotone krow, in fast-mode
    # this slips through:
    wateroil.table.loc[0:2, "KROW"] = [1.00, 0.81, 0.81]
    assert "0.1000000 0.0100000 0.8100000 0.0000000" in wateroil.SWOF()
    assert "0.2000000 0.0400000 0.8100000 0.0000000" in wateroil.SWOF()
    # not corrected:               ^^^^^^

    wateroil.table.loc[0:2, "KRW"] = [0.00, 0.01, 0.01]
    assert "0.1000000 0.0100000" in wateroil.SWFN()
    assert "0.2000000 0.0100000" in wateroil.SWFN()


def test_swfn():
    """Test that we can dump SWFN without giving oil relperm"""
    wateroil = WaterOil(h=0.1)
    wateroil.add_corey_water()
    swfnstr = wateroil.SWFN()
    assert "SWFN" in swfnstr
    assert len(swfnstr) > 15


def test_linearsegments():
    """Made for testing the linear segments during
    the resolution of issue #163"""
    wateroil = WaterOil(h=0.01, swl=0.1, swcr=0.3, sorw=0.3)
    wateroil.add_corey_oil(now=10, kroend=0.5)
    wateroil.add_corey_water(nw=10, krwend=0.5)
    check_table(wateroil.table)
    check_linear_sections(wateroil)


@pytest.mark.skipif(sys.platform == "win32", reason="Plotting not stable on windows")
def test_plotting(mocker):
    """Test that plotting code pass through (nothing displayed)"""
    mocker.patch("matplotlib.pyplot.show", return_value=None)
    wateroil = WaterOil(swl=0.1, h=0.1)
    wateroil.add_corey_water()
    wateroil.add_corey_oil()
    wateroil.plotkrwkrow(mpl_ax=matplotlib.pyplot.subplots()[1])
    wateroil.plotkrwkrow(logyscale=True, mpl_ax=matplotlib.pyplot.subplots()[1])
    wateroil.plotkrwkrow(mpl_ax=None)

    wateroil.add_simple_J()
    wateroil.plotpc(mpl_ax=matplotlib.pyplot.subplots()[1])
    wateroil.plotpc(mpl_ax=None)
    wateroil.plotpc(logyscale=True)


def test_wateroil_linear():
    """Test linear wateroil curves"""
    wateroil = WaterOil(h=1)
    wateroil.add_corey_water()
    wateroil.add_corey_oil()
    swofstr = wateroil.SWOF(header=False)
    check_table(wateroil.table)
    check_linear_sections(wateroil)
    assert isinstance(swofstr, str)
    assert swofstr
    assert len(wateroil.table) == 2
    assert np.isclose(wateroil.crosspoint(), 0.5)

    # What if there is no space for our choice of h?
    # We should be able to initialize nonetheless
    # (a warning could be given)
    wateroil = WaterOil(swl=0.1, h=1)
    wateroil.add_corey_water()
    wateroil.add_corey_oil()
    check_table(wateroil.table)
    check_linear_sections(wateroil)
    assert len(wateroil.table) == 2
    assert np.isclose(wateroil.table["SW"].min(), 0.1)
    assert np.isclose(wateroil.table["SW"].max(), 1.0)
    assert np.isclose(wateroil.crosspoint(), 0.55)


def test_comments():
    """Test that the outputters include endpoints in comments"""
    wateroil = WaterOil(h=0.3)
    wateroil.add_corey_water()
    wateroil.add_corey_oil()
    swfn = wateroil.SWFN()
    assert "--" in swfn
    assert "pyscal: " in swfn  # part of version string
    assert "swirr=0" in swfn
    assert "swcr=0" in swfn
    assert "swl=0" in swfn
    assert "sorw=0" in swfn
    assert "socr=0" not in swfn  # Only included when nonzero
    assert "nw=2" in swfn
    assert "krwend=1" in swfn
    assert "Corey" in swfn
    assert "krw = krow @ sw=0.5" in swfn
    assert "Zero capillary pressure" in swfn
    assert "SW" in swfn
    assert "KRW" in swfn
    assert "KROW" not in swfn
    assert "PC" in swfn

    swof = wateroil.SWOF()
    assert "--" in swof
    assert "pyscal: " in swof  # part of version string
    assert "swirr=0" in swof
    assert "swcr=0" in swof
    assert "swl=0" in swof
    assert "sorw=0" in swof
    assert "socr=0" not in swof  # Only included when nonzero
    assert "nw=2" in swof
    assert "now=2" in swof
    assert "krwend=1" in swof
    assert "Corey" in swof
    assert "krw = krow @ sw=0.5" in swof
    assert "Zero capillary pressure" in swof
    assert "SW" in swof
    assert "KRW" in swof
    assert "KROW" in swof
    assert "PC" in swof

    paleo_wo = WaterOil(h=0.3, socr=0.1)
    paleo_wo.add_corey_water()
    paleo_wo.add_corey_oil()
    assert "socr=0.1" in paleo_wo.SWFN()
    assert "socr=0.1" in paleo_wo.SWOF()


def test_nexus():
    """Test the Nexus export"""
    wateroil = WaterOil(h=0.01, swl=0.1, swcr=0.3, sorw=0.3)
    wateroil.add_corey_oil(now=10, kroend=0.5)
    wateroil.add_corey_water(nw=10, krwend=0.5)
    nexus_lines = wateroil.WOTABLE().splitlines()
    no_comments = [line for line in nexus_lines if not line.startswith("!") or not line]
    assert no_comments[0] == "WOTABLE"
    assert no_comments[1] == "SW KRW KROW PC"
    df = pd.read_table(
        io.StringIO("\n".join(no_comments[2:])),
        engine="python",
        sep=r"\s+",
        header=None,
    )
    assert (df.to_numpy() <= 1.0).all()
    assert (df.to_numpy() >= 0.0).all()


@pytest.mark.parametrize(
    "columnname, errorvalues",
    [
        ("SW", [1, 0]),
        ("SW", [0, 2]),
        ("KRW", [1, 0]),
        ("KRW", [0, 2]),
        ("KROW", [0, 1]),
        ("KROW", [2, 0]),
        ("PC", [np.inf, 0]),
        ("PC", [np.nan, 0]),
        ("PC", [1, 2]),
        ("PC", [0, 1]),
    ],
)
def test_selfcheck(columnname, errorvalues):
    """Test the selfcheck feature of a WaterOil object"""
    wateroil = WaterOil(h=1)
    wateroil.add_corey_water()
    wateroil.add_corey_oil()
    assert wateroil.selfcheck()

    # Punch the internal table directly to trigger error:
    wateroil.table[columnname] = errorvalues
    assert not wateroil.selfcheck()
    assert wateroil.SWOF() == ""
    if columnname != "KROW":
        assert wateroil.SWFN() == ""



"""Test the pyscal client"""

import logging
import os
import subprocess
import sys
from pathlib import Path

import pandas as pd
import pytest

from pyscal import pyscalcli
from pyscal.utils.testing import sat_table_str_ok


@pytest.mark.integration
def test_installed():
    """Test that the command line client is installed in PATH and
    starts up nicely"""
    assert subprocess.check_output(["pyscal", "-h"])


@pytest.mark.skipif(sys.version_info < (3, 7), reason="Requires Python 3.7 or higher")
@pytest.mark.parametrize("verbosity_flag", [None, "--verbose", "--debug"])
def test_log_levels(tmp_path, verbosity_flag):
    """Test that we can control the log level from the command line
    client, and get log output from modules deep down"""

    relperm_file = str(
        # A cell in this xlsx contains "Åre 1.5", does not work on Windows
        Path(__file__).absolute().parent / "data" / "relperm-input-example.xlsx"
    )

    commands = ["pyscal", relperm_file]
    if verbosity_flag is not None:
        commands.append(verbosity_flag)

    result = subprocess.run(commands, cwd=tmp_path, capture_output=True, check=True)
    output = result.stdout.decode() + result.stderr.decode()

    if verbosity_flag is None:
        assert "INFO:" not in output
        assert "DEBUG:" not in output
    elif verbosity_flag == "--verbose":
        assert "INFO:" in output
        assert "DEBUG:" not in output
        assert "Loaded input data" in output
        assert "Keywords SWOF, SGOF (family 1) for 3 SATNUMs generated" in output
    elif verbosity_flag == "--debug":
        assert "INFO:" in output
        assert "DEBUG:" in output
        assert "Initialized GasOil with" in output
        assert "Initialized WaterOil with" in output
    else:
        raise ValueError("Unknown value for 'verbosity_flag'")


def test_pyscal_client_static(tmp_path, caplog, default_loglevel, mocker):
    # default_loglevel fixture is in conftest.py
    """Test pyscal client for static relperm input"""
    testdir = Path(__file__).absolute().parent
    relperm_file = testdir / "data/relperm-input-example.xlsx"

    os.chdir(tmp_path)

    caplog.clear()
    mocker.patch("sys.argv", ["pyscal", str(relperm_file)])
    pyscalcli.main()
    assert Path("relperm.inc").is_file()
    assert not any(record.levelno == logging.ERROR for record in caplog.records)
    assert not any(record.levelno == logging.INFO for record in caplog.records)

    relpermlines = os.linesep.join(
        Path("relperm.inc").read_text(encoding="utf8").splitlines()
    )
    assert "SWOF" in relpermlines
    assert "SGOF" in relpermlines
    assert "SLGOF" not in relpermlines
    assert "SOF3" not in relpermlines
    sat_table_str_ok(relpermlines)

    caplog.clear()
    mocker.patch(
        "sys.argv", ["pyscal", str(relperm_file), "--output", "alt2relperm.inc"]
    )
    pyscalcli.main()
    assert Path("alt2relperm.inc").is_file()
    assert not any(record.levelno == logging.ERROR for record in caplog.records)

    caplog.clear()
    mocker.patch("sys.argv", ["pyscal", str(relperm_file), "-o", "altrelperm.inc"])
    pyscalcli.main()
    assert Path("altrelperm.inc").is_file()
    assert not any(record.levelno == logging.ERROR for record in caplog.records)

    caplog.clear()
    mocker.patch(
        "sys.argv", ["pyscal", str(relperm_file), "--family2", "-o", "relperm-fam2.inc"]
    )
    pyscalcli.main()
    assert not any(record.levelno == logging.ERROR for record in caplog.records)
    assert Path("relperm-fam2.inc").is_file()
    relpermlines = os.linesep.join(
        Path("relperm-fam2.inc").read_text(encoding="utf8").splitlines()
    )
    assert "SWFN" in relpermlines
    assert "SGFN" in relpermlines
    assert "SOF3" in relpermlines
    assert "SWOF" not in relpermlines
    assert "SGOF" not in relpermlines
    sat_table_str_ok(relpermlines)

    caplog.clear()
    mocker.patch(
        "sys.argv",
        ["pyscal", str(relperm_file), "--slgof", "--output", "relperm-slgof.inc"],
    )
    pyscalcli.main()
    assert not any(record.levelno == logging.ERROR for record in caplog.records)
    assert Path("relperm-slgof.inc").is_file()
    relpermlines = os.linesep.join(
        Path("relperm-slgof.inc").read_text(encoding="utf8").splitlines()
    )
    assert "SWOF" in relpermlines
    assert "SGOF" not in relpermlines
    assert "SLGOF" in relpermlines
    assert "SOF3" not in relpermlines
    sat_table_str_ok(relpermlines)

    caplog.clear()
    # Dump to deep directory structure that does not exists
    mocker.patch(
        "sys.argv",
        [
            "pyscal",
            str(relperm_file),
            "--family2",
            "-o",
            "eclipse/include/props/relperm-fam2.inc",
        ],
    )
    with pytest.raises(SystemExit, match="Output directory not found"):
        pyscalcli.main()
    assert not Path("eclipse/include/props/relperm-fam2.inc").is_file()

    caplog.clear()
    mocker.patch(
        "sys.argv", ["pyscal", str(relperm_file), "-o", "include/props/relperm.inc"]
    )
    with pytest.raises(SystemExit, match="Output directory not found"):
        pyscalcli.main()
    assert not Path("include/props/relperm.inc").is_file()

    caplog.clear()
    # Check that we can read specific sheets
    mocker.patch(
        "sys.argv",
        [
            "pyscal",
            str(relperm_file),
            "--sheet_name",
            "relperm",
            "--output",
            "relperm-firstsheet.inc",
        ],
    )
    pyscalcli.main()
    assert not any(record.levelno == logging.ERROR for record in caplog.records)

    # Identical files:
    assert len(
        Path("relperm-firstsheet.inc").read_text(encoding="utf8").splitlines()
    ) == len(Path("relperm.inc").read_text(encoding="utf8").splitlines())

    # Check that we can read specific sheets
    caplog.clear()
    mocker.patch(
        "sys.argv",
        [
            "pyscal",
            str(relperm_file),
            "--sheet_name",
            "simple",
            "--output",
            "relperm-secondsheet.inc",
        ],
    )
    pyscalcli.main()
    assert not any(record.levelno == logging.ERROR for record in caplog.records)
    secondsheet = os.linesep.join(
        Path("relperm-secondsheet.inc").read_text(encoding="utf8").splitlines()
    )
    assert "SATNUM 3" not in secondsheet
    assert "sand" in secondsheet
    assert "mud" in secondsheet  # From the comment column in sheet: simple

    # Check that we can read specific sheets
    caplog.clear()
    mocker.patch(
        "sys.argv",
        [
            "pyscal",
            str(relperm_file),
            "--sheet_name",
            "NOTEXISTINGÆÅ",
            "--output",
            "relperm-empty.inc",
        ],
    )
    with pytest.raises(SystemExit):
        pyscalcli.main()
    assert not Path("relperm-empty.inc").is_file()

    caplog.clear()
    mocker.patch(
        "sys.argv",
        ["pyscal", str(relperm_file), "--delta_s", "0.1", "-o", "deltas0p1.inc"],
    )
    pyscalcli.main()
    assert not any(record.levelno == logging.ERROR for record in caplog.records)
    linecount1 = len(Path("deltas0p1.inc").read_text(encoding="utf8").splitlines())

    caplog.clear()
    mocker.patch(
        "sys.argv",
        [
            "pyscal",
            str(relperm_file),
            "--delta_s",
            "0.01",
            "-o",
            "deltas0p01.inc",
        ],
    )
    pyscalcli.main()
    assert not any(record.levelno == logging.ERROR for record in caplog.records)
    linecount2 = len(Path("deltas0p01.inc").read_text(encoding="utf8").splitlines())
    assert linecount2 > linecount1 * 4  # since we don't filter out non-numerical lines


def test_pyscalcli_stdout_output(capsys, mocker):
    """Test that we can write to stdout"""
    scalrec_file = Path(__file__).absolute().parent / "data/scal-pc-input-example.xlsx"
    mocker.patch(
        "sys.argv",
        ["pyscal", str(scalrec_file), "--int_param_wo", "0", "--output", "-"],
    )
    pyscalcli.main()
    captured = capsys.readouterr()
    assert "SWOF" in captured.out

    mocker.patch(
        "sys.argv",
        [
            "pyscal",
            str(scalrec_file),
            "--family2",
            "--int_param_wo",
            "0",
            "--output",
            "-",
        ],
    )
    pyscalcli.main()
    captured = capsys.readouterr()
    assert "SOF3" in captured.out


def test_pyscalcli_let_pd_and_imb(capsys, mocker):
    """Test that we can write to stdout"""
    scalrec_file = (
        Path(__file__).absolute().parent / "data/pyscal_config_simple_let_hyst.csv"
    )
    mocker.patch(
        "sys.argv",
        ["pyscal", str(scalrec_file), "--output", "-"],
    )
    pyscalcli.main()
    captured = capsys.readouterr()
    assert "SWOF" in captured.out
    assert "SGOF" in captured.out
    assert "LET correlation for primary drainage Pc" in captured.out
    assert "LET correlation for imbibition Pc" in captured.out


def test_pyscalcli_pc_skjaeveland(capsys, mocker):
    """Test that we can write to stdout"""
    scalrec_file = (
        Path(__file__).absolute().parent / "data/relperm-input-example_skjaeveland.csv"
    )
    mocker.patch(
        "sys.argv",
        ["pyscal", str(scalrec_file), "--output", "-"],
    )
    pyscalcli.main()
    captured = capsys.readouterr()
    assert "SWOF" in captured.out
    assert "SGOF" in captured.out
    assert "Skjæveland correlation for Pc" in captured.out


def test_pyscalcli_exception_catching(capsys, mocker):
    """The command line client catches selected exceptions.

    Traceback is always included."""
    mocker.patch("sys.argv", ["pyscal", "notexisting.xlsx"])
    with pytest.raises(SystemExit, match="File not found"):
        pyscalcli.main()
    outerr = capsys.readouterr().out + capsys.readouterr().err
    assert "raise" in outerr  # This is the traceback.


def test_pyscalcli_oilwater(tmp_path, caplog, mocker):
    """Test the command line client in two-phase oil-water"""
    os.chdir(tmp_path)
    relperm_file = "oilwater.csv"
    pd.DataFrame(
        # "fooå" here causes problems on Windows
        columns=["SATNUM", "nw", "now", "tag"],
        data=[[1, 2, 3, "fooå"]],
    ).to_csv(relperm_file, index=False, encoding="utf-8-sig")
    caplog.clear()
    mocker.patch(
        "sys.argv",
        [
            "pyscal",
            relperm_file,
            "--output",
            "ow.inc",
        ],
    )
    pyscalcli.main()
    assert not any(record.levelno == logging.WARNING for record in caplog.records)
    assert not any(record.levelno == logging.ERROR for record in caplog.records)
    lines = Path("ow.inc").read_text(encoding="utf8").splitlines()
    joined = os.linesep.join(lines)
    assert "fooå" in joined
    assert 100 < len(lines) < 120  # weak test..

    # Test with SCAL recommendation:
    pd.DataFrame(
        columns=["SATNUM", "case", "nw", "now", "tag"],
        data=[
            [1, "low", 2, 3, "fooå"],
            [1, "base", 2, 3, "fooå"],
            [1, "high", 2, 3, "fooå"],
        ],
    ).to_csv(relperm_file, index=False)
    caplog.clear()
    mocker.patch(
        "sys.argv",
        [
            "pyscal",
            relperm_file,
            "--int_param_wo",
            "-0.1",
            "--output",
            "ow-int.inc",
        ],
    )
    pyscalcli.main()
    assert not any(record.levelno == logging.WARNING for record in caplog.records)
    assert not any(record.levelno == logging.ERROR for record in caplog.records)


def test_pyscalcli_gaswater(tmp_path, caplog, mocker):
    """Test the command line endpoint on gas-water problems"""
    os.chdir(tmp_path)
    relperm_file = "gaswater.csv"
    pd.DataFrame(columns=["SATNUM", "nw", "ng"], data=[[1, 2, 3]]).to_csv(
        relperm_file, index=False
    )
    caplog.clear()
    mocker.patch(
        "sys.argv",
        [
            "pyscal",
            relperm_file,
            "--output",
            "gw.inc",
        ],
    )
    pyscalcli.main()
    assert not any(record.levelno == logging.WARNING for record in caplog.records)
    assert not any(record.levelno == logging.ERROR for record in caplog.records)
    lines = Path("gw.inc").read_text(encoding="utf8").splitlines()
    joined = os.linesep.join(lines)
    assert "SWFN" in joined
    assert "SGFN" in joined
    assert "SWOF" not in joined
    assert "sgrw" in joined
    assert "krgendanchor" not in joined
    assert "sorw" not in joined
    assert "sorg" not in joined
    assert len(lines) > 40


def test_pyscalcli_gaswater_scal(tmp_path, caplog, mocker):
    """Test the command line endpoint on gas-water problems, with
    interpolation"""
    os.chdir(tmp_path)
    relperm_file = "gaswater.csv"
    pd.DataFrame(
        columns=["SATNUM", "CASE", "nw", "ng"],
        data=[[1, "pess", 2, 3], [1, "base", 3, 4], [1, "opt", 5, 6]],
    ).to_csv(relperm_file, index=False)

    caplog.clear()
    mocker.patch(
        "sys.argv",
        [
            "pyscal",
            relperm_file,
            "--int_param_wo",
            "-0.2",
            "--output",
            "gw.inc",
        ],
    )
    pyscalcli.main()
    assert not any(record.levelno == logging.INFO for record in caplog.records)
    assert not any(record.levelno == logging.WARNING for record in caplog.records)
    assert not any(record.levelno == logging.ERROR for record in caplog.records)
    lines = Path("gw.inc").read_text(encoding="utf8").splitlines()
    joined = os.linesep.join(lines)
    assert "SWFN" in joined
    assert "SGFN" in joined
    assert "SWOF" not in joined
    assert "sgrw" in joined
    assert "krgendanchor" not in joined
    assert "sorw" not in joined
    assert "sorg" not in joined
    assert len(lines) > 40


def test_pyscal_client_scal(tmp_path, caplog, default_loglevel, mocker):
    """Test the command line endpoint on SCAL recommendation"""
    scalrec_file = Path(__file__).absolute().parent / "data/scal-pc-input-example.xlsx"

    os.chdir(tmp_path)

    mocker.patch("sys.argv", ["pyscal", str(scalrec_file)])
    with pytest.raises(SystemExit):
        pyscalcli.main()

    caplog.clear()
    mocker.patch(
        "sys.argv",
        ["pyscal", str(scalrec_file), "--int_param_wo", 0, "-o", "relperm1.inc"],
    )
    pyscalcli.main()
    assert not any(record.levelno == logging.INFO for record in caplog.records)
    assert not any(record.levelno == logging.WARNING for record in caplog.records)
    assert not any(record.levelno == logging.ERROR for record in caplog.records)

    relpermlines = os.linesep.join(
        Path("relperm1.inc").read_text(encoding="utf8").splitlines()
    )
    assert "SWOF" in relpermlines
    assert "SGOF" in relpermlines
    assert "SLGOF" not in relpermlines
    assert "SOF3" not in relpermlines
    sat_table_str_ok(relpermlines)
    # assert "int_param_wo: 0\n" in relpermlines  # this should be in the tag.

    caplog.clear()
    mocker.patch(
        "sys.argv",
        [
            "pyscal",
            str(scalrec_file),
            "--int_param_wo",
            "-0.5",
            "-o",
            "relperm2.inc",
        ],
    )
    pyscalcli.main()
    assert not any(record.levelno == logging.INFO for record in caplog.records)
    assert not any(record.levelno == logging.WARNING for record in caplog.records)
    assert not any(record.levelno == logging.ERROR for record in caplog.records)
    # assert something about -0.5 in the comments


def test_pyscal_client_error(tmp_path, mocker):
    """Test various error conditions, asserting the the correct error message is emitted

    Some error are caught in pyscalcli.py, some errors are caught when loading the xlsx
    file"""

    os.chdir(tmp_path)
    scalrec_file = Path(__file__).absolute().parent / "data/scal-pc-input-example.xlsx"

    # int_param_go should not be used alone:
    mocker.patch("sys.argv", ["pyscal", str(scalrec_file), "--int_param_go", "-0.5"])
    with pytest.raises(SystemExit, match="Don't use int_param_go alone"):
        pyscalcli.main()

    # Delete SATNUM from xlsx input:
    pd.read_excel(scalrec_file, engine="openpyxl").drop("SATNUM", axis=1).to_csv(
        "no_satnum.csv"
    )
    mocker.patch("sys.argv", ["pyscal", "no_satnum.csv", "--int_param_wo", "-0.5"])
    with pytest.raises(SystemExit, match="SATNUM must be present"):
        pyscalcli.main()

    # Delete CASE from xlsx input:
    pd.read_excel(scalrec_file, engine="openpyxl").drop("CASE", axis=1).to_csv(
        "no_case.csv"
    )
    mocker.patch("sys.argv", ["pyscal", "no_case.csv", "--int_param_wo", "-0.5"])
    with pytest.raises(SystemExit, match="Non-unique SATNUMs"):
        pyscalcli.main()

    # Multiple interpolation parameters, this was supported in pyscal <= 0.7.7,
    # but is now an error (raised by argparse):
    mocker.patch(
        "sys.argv", ["pyscal", str(scalrec_file), "--int_param_wo", "-0.5", "0"]
    )
    with pytest.raises(SystemExit):
        pyscalcli.main()


@pytest.mark.parametrize(
    "int_param_wo, int_param_go, raises, match",
    [
        (-1, None, None, None),
        (-2, None, ValueError, "must be in"),
        (-1, -2, ValueError, "must be in"),
        (None, -1, ValueError, "int_param_go alone"),
        (None, None, ValueError, None),
        ([], None, TypeError, "SATNUM specific interpolation"),
        ([-1], None, TypeError, "SATNUM specific interpolation"),
        ([1, 1], None, TypeError, "SATNUM specific interpolation"),
        ([], [], TypeError, "SATNUM specific interpolation"),
        (-1, [], TypeError, "SATNUM specific interpolation"),
        (None, [], ValueError, "int_param_go alone"),
    ],
)
def test_pyscal_main_interpolation_parameters(
    int_param_wo, int_param_go, raises, match
):
    """Define the behaviour on different interpolation parameter combinations.

    Earlier pyscal versions allowed lists of parameters as input.

    The command line client will also catch these errors through argparse, but
    when pyscal is used e.g. in fm_pyscal.py in semeio, these errors need to
    be caught by the main() function.
    """
    scalrec_file = Path(__file__).absolute().parent / "data/scal-pc-input-example.xlsx"
    if raises is not None:
        with pytest.raises(raises, match=match):
            pyscalcli.pyscal_main(
                scalrec_file,
                int_param_wo=int_param_wo,
                int_param_go=int_param_go,
                output=os.devnull,
            )
    else:
        pyscalcli.pyscal_main(
            scalrec_file,
            int_param_wo=int_param_wo,
            int_param_go=int_param_go,
            output=os.devnull,
        )


def test_pyscal_main():
    """The pyscal client is a wrapper main() function that runs argparse, and then
    hands over responsibility to pyscal_main(). This wrapping is to facilitate
    fm_pyscal.py in semeio f.ex.

    This test function is for testing e.g error scenarios that argparse
    would catch, but that we also need to check on behalf of semeio usage."""

    # This input file has no CASE column, and interpolation is not meaningful.
    relperm_file = Path(__file__).absolute().parent / "data/relperm-input-example.xlsx"

    pyscalcli.pyscal_main(relperm_file, output=os.devnull)

    with pytest.raises(ValueError, match="Interpolation parameter provided"):
        pyscalcli.pyscal_main(relperm_file, int_param_wo=-1, output=os.devnull)


def test_pyscalcli_plot(capsys, mocker, tmpdir):
    """Test that plots are created through the CLI. This is done by testing to
    see if the print statements in the save_figure function are present in stdout"""
    scalrec_file = Path(__file__).absolute().parent / "data/scal-pc-input-example.xlsx"

    mocker.patch(
        "sys.argv",
        [
            "pyscal",
            str(scalrec_file),
            "--int_param_wo",
            "0",
            "--output",
            "-",
            "--plot",
            "--plot_pc",
            "--plot_outdir",
            str(tmpdir),
        ],
    )

    pyscalcli.main()

    expected_plots = [
        "krw_krow_SATNUM_1.png",
        "krg_krog_SATNUM_1.png",
        "krw_krow_SATNUM_2.png",
        "krg_krog_SATNUM_2.png",
        "krw_krow_SATNUM_3.png",
        "krg_krog_SATNUM_3.png",
        "pcow_SATNUM_1.png",
        "pcow_SATNUM_2.png",
        "pcow_SATNUM_3.png",
    ]

    captured = capsys.readouterr()

    for plot in expected_plots:
        assert plot in captured.out



"""Test the pyscal module to be used for testing.

These functions also need to be correct in order to trust tests."""

from contextlib import contextmanager

import pandas as pd
import pytest

from pyscal.utils.testing import float_df_checker, sat_table_str_ok


@contextmanager
def does_not_raise():
    yield


@pytest.mark.parametrize(
    "string, expectation",
    [
        pytest.param(None, pytest.raises(AssertionError)),
        pytest.param("", pytest.raises(AssertionError)),
        pytest.param(
            "SWOF\n 0 0 0\n/",
            pytest.raises(AssertionError),
            id="missing_pyscal_header",
        ),
        ("-- pyscal: \nSWOF\n0 0 0 0 \n/", does_not_raise()),
        pytest.param(
            "-- pyscal: \n\nSWOF\n0 0 0 0 \n/",
            does_not_raise(),
            id="whitespace_suffix_allowed",
        ),
        pytest.param(
            "-- pyscal: \n\nSWOF\n0 0 0\n/", does_not_raise(), id="three_numbers_ok"
        ),
        pytest.param(
            "-- pyscal: \n\nSWOF\n0 0\n/",
            pytest.raises(AssertionError),
            id="two_numbers_not_ok",
        ),
        ("-- pyscal: \nSGOF\n0 0 0 0 \n/", does_not_raise()),
        ("-- pyscal: \n-- foo\nSGOF\n0 0 0 0 \n/", does_not_raise()),
        pytest.param(
            "-- pyscal: \nRWOF\n0 0 0 0 \n/",
            pytest.raises(AssertionError),
            id="invalid_keyword",
        ),
        pytest.param(
            "-- pyscal: \nSROF\n0 0 0 0 \n/",
            does_not_raise(),
            id="invalid_keyword_that_passes",
        ),
        pytest.param(
            "-- pyscal: \nSWOF\n 0 0 0 0 \n/",
            pytest.raises(AssertionError),
            id="whitespace_prefix_disallowed",
        ),
        pytest.param(
            "-- pyscal: \nSWOF\na 0 0 0 \n/",
            pytest.raises(AssertionError),
            id="not_float",
        ),
        pytest.param(
            "-- pyscal: \nSWOF\n0 0 0 0 0\n/",
            pytest.raises(AssertionError),
            id="too_many_floats",
        ),
        pytest.param(
            "-- pyscal: \nSWOF\n-1 0 0 0\n/",
            pytest.raises(AssertionError),
            id="negative_number",
        ),
        pytest.param(
            "-- pyscal: \nSWOF\n0 0 -1 0\n/",
            pytest.raises(AssertionError),
            id="negative_number",
        ),
        pytest.param(
            "-- pyscal: \nSWOF\n0 0 -0 0\n/",
            pytest.raises(AssertionError),
            id="negative_zero",
        ),
        pytest.param(
            "-- pyscal: \nSWOF\n1.001 0 0 0\n/",
            pytest.raises(AssertionError),
            id="saturation_above_1",
        ),
        pytest.param(
            "-- pyscal: \nSWOF\n1 0 0 20.000000\n/",
            does_not_raise(),
            id="cap_pressure_can_be_larger",
        ),
        pytest.param(
            "-- pyscal: \nSWOF\n0 0 0 0\n1 1 1\n/",
            pytest.raises(AssertionError),
            id="non_rectangular_frame",
        ),
    ],
)
def test_sat_table_str_ok(string, expectation):
    """Test that we can determine whether a string is accepted by Eclipse

    (this is instead of actually testing in the simulator)
    """
    with expectation:
        sat_table_str_ok(string)


@pytest.mark.parametrize(
    "data, value, expected, expectation",
    [
        ([[0, 1], [1, 2]], 0, 1, does_not_raise()),
        pytest.param([[0, 1], [1, 2]], 0, 2, pytest.raises(AssertionError)),
        ([[0, 1], [1, 2]], 1, 2, does_not_raise()),
        ([[1e-07, 1], [1, 2]], 0, 1, does_not_raise()),
        ([[1e-06, 1], [1, 2]], 0, 1, does_not_raise()),
        ([[1e-05, 1], [1, 2]], 0, 1, does_not_raise()),
        ([[0.0001, 1], [1, 2]], 0, 1, does_not_raise()),
        ([[0.001, 1], [1, 2]], 0, 1, does_not_raise()),
        ([[0.01, 1], [1, 2]], 0, 1, does_not_raise()),
        ([[0.1, 1], [1, 2]], 0, 1, does_not_raise()),
        ([[-0.1, 1], [1, 2]], 0, 1, does_not_raise()),
        ([[-0.001, 1], [1, 2]], 0, 1, does_not_raise()),
        ([[0, 1], [0.1, 1.5], [1, 2]], 0.1, 1.5, does_not_raise()),
        ([[0.09, 1], [0.1, 1.5], [1, 2]], 0.1, 1.5, does_not_raise()),
        ([[0.099, 1], [0.1, 1.5], [1, 2]], 0.1, 1.5, does_not_raise()),
        ([[0.0999, 1], [0.1, 1.5], [1, 2]], 0.1, 1.5, does_not_raise()),
        ([[0.09999, 1], [0.1, 1.5], [1, 2]], 0.1, 1.5, does_not_raise()),
        ([[0.099999, 1], [0.1, 1.5], [1, 2]], 0.1, 1.5, does_not_raise()),
        ([[0.09999999, 1], [0.1, 1.5], [1, 2]], 0.1, 1.5, does_not_raise()),
        pytest.param([[0, 0], [1, 1]], 0, 1e-05, does_not_raise()),
        pytest.param(
            [[0, 0], [1, 1]],
            0,
            0.0001,
            pytest.raises(AssertionError),
            id="too_far_from_zero",
        ),
        pytest.param([[0, 0], [1, 1]], 1, 1.000001, does_not_raise()),
        pytest.param(
            [[0, 0], [1, 1]],
            1,
            1.0001,
            pytest.raises(AssertionError),
            id="too_far_from_one",
        ),
    ],
)
def test_float_df_checker(data, value, expected, expectation):
    """Test that we can lookup in a dataframe for values and test them"""
    with expectation:
        assert float_df_checker(
            pd.DataFrame(columns=["idx", "values"], data=data),
            "idx",
            value,
            "values",
            expected,
        )



"""Test module for the GasWater object"""

import sys

import hypothesis.strategies as st
import matplotlib
import matplotlib.pyplot
import numpy as np
import pytest
from hypothesis import given

from pyscal import GasWater
from pyscal.constants import SWINTEGERS
from pyscal.utils.testing import (
    check_linear_sections,
    check_table,
    float_df_checker,
    sat_table_str_ok,
)


def check_endpoints(gaswater, krwend, krwmax, krgend):
    """Check that the code produces correct endpoints for
    parametrizations, on discrete cases"""
    swtol = 10 / SWINTEGERS

    # Check endpoints for gas curve:
    # krg at swl should be krgend:
    assert float_df_checker(gaswater.gasoil.table, "SGN", 1.0, "KRG", krgend)
    assert float_df_checker(gaswater.gasoil.table, "SL", gaswater.swl, "KRG", krgend)
    # krg at sgcr (sgn is zero there) should be zero:
    assert float_df_checker(gaswater.gasoil.table, "SGN", 0.0, "KRG", 0.0)
    assert float_df_checker(gaswater.gasoil.table, "SL", 1 - gaswater.sgcr, "KRG", 0.0)

    check_linear_sections(gaswater.gasoil)
    check_linear_sections(gaswater.wateroil)

    # Check endpoints for water curve: (np.isclose is only reliable around 1)
    assert float_df_checker(gaswater.wateroil.table, "SWN", 0.0, "KRW", 0.0)
    assert float_df_checker(gaswater.wateroil.table, "SW", gaswater.swcr, "KRW", 0)

    if gaswater.sgrw > swtol:
        # (hard to get it right when sgrw is less than h and close to zero)
        assert float_df_checker(
            gaswater.wateroil.table, "SW", 1 - gaswater.sgrw, "KRW", krwend
        )
        assert np.isclose(gaswater.wateroil.table["KRW"].max(), krwmax)
    else:
        assert np.isclose(gaswater.wateroil.table["KRW"].max(), krwend)


def test_constructor():
    """Test that object attributes are set upon intialization"""
    gaswater = GasWater()
    gaswater.add_corey_gas()
    gaswater.add_corey_water()
    assert "Corey" in gaswater.krwcomment
    assert "Corey" in gaswater.krgcomment

    # Trigger tag inconsistency:
    gaswater.wateroil.tag = "Foo"
    gaswater.gasoil.tag = "Bar"
    with pytest.raises(ValueError, match="Internal tag-inconsistency in GasWater"):
        gaswater.tag

    # Trigger self-check bug warning:
    gaswater.gasoil = None
    assert gaswater.selfcheck() is False


def test_fast():
    """Test the fast mode, skipping some computations"""
    gaswater = GasWater(fast=True)
    assert gaswater.fast
    gaswater.add_corey_gas()
    gaswater.add_corey_water()
    swfn = gaswater.SWFN()
    assert "krw = krg" not in swfn  # Crosspoint should not be present
    sgfn = gaswater.SGFN()
    assert "krw = krg" not in sgfn  # Crosspoint should not be present


@given(st.text())
def test_gaswater_tag(tag):
    """Test that we are unlikely to crash Eclipse
    by having ugly tag names"""
    gaswater = GasWater(h=0.5, tag=tag)
    gaswater.add_corey_gas()
    gaswater.add_corey_water()
    sat_table_str_ok(gaswater.SWFN())
    sat_table_str_ok(gaswater.SGFN())


@given(st.floats(), st.floats())
def test_gaswater_corey1(nw, ng):
    """Test random corey parameters"""
    gaswater = GasWater()
    try:
        gaswater.add_corey_gas(ng=ng)
        gaswater.add_corey_water(nw=nw)
    except AssertionError:
        # This happens for "invalid" input
        return

    assert "KRG" in gaswater.gasoil.table
    assert "KRW" in gaswater.wateroil.table
    assert isinstance(gaswater.wateroil.krwcomment, str)
    check_table(gaswater.wateroil.table)
    check_table(gaswater.gasoil.table)
    check_linear_sections(gaswater.wateroil)
    check_linear_sections(gaswater.gasoil)
    swfnstr = gaswater.SWFN()
    assert len(swfnstr) > 100
    sgfnstr = gaswater.SGFN()
    assert len(sgfnstr) > 100


@given(st.floats(), st.floats(), st.floats(), st.floats(), st.floats())
def test_gaswater_let1(l, e, t, krwend, krwmax):
    """Test random LET parameters"""
    gaswater = GasWater()
    try:
        gaswater.add_LET_gas(l, e, t, krwend)
        gaswater.add_LET_water(l, e, t, krwend, krwmax)
    except AssertionError:
        # This happens for negative values f.ex.
        return
    assert "KRG" in gaswater.gasoil.table
    assert "KRW" in gaswater.wateroil.table
    assert isinstance(gaswater.wateroil.krwcomment, str)
    check_table(gaswater.wateroil.table)
    check_table(gaswater.gasoil.table)
    check_linear_sections(gaswater.wateroil)
    check_linear_sections(gaswater.gasoil)
    swfnstr = gaswater.SWFN()
    assert len(swfnstr) > 100
    sgfnstr = gaswater.SGFN()
    assert len(sgfnstr) > 100


@given(
    st.floats(min_value=0, max_value=0.3),
    st.floats(min_value=0, max_value=0.3),
    st.floats(min_value=0, max_value=0.3),
    st.floats(min_value=0.1, max_value=1),
    st.floats(min_value=0.1, max_value=1),
    st.floats(min_value=0.1, max_value=1),
    st.floats(min_value=0.001, max_value=1),
    st.booleans(),
)
def test_gaswater_krendmax(swl, swcr, sgrw, krgend, krwend, krwmax, h, fast):
    """Test endpoints for gaswater using hypothesis testing"""
    try:
        gaswater = GasWater(swl=swl, swcr=swcr, sgrw=sgrw, h=h, fast=fast)
    except AssertionError:
        return
    krwend = min(krwend, krwmax)
    gaswater.add_corey_gas(krgend=krgend)
    gaswater.add_corey_water(krwend=krwend, krwmax=krwmax)
    check_table(gaswater.wateroil.table)
    check_table(gaswater.gasoil.table)
    assert gaswater.selfcheck()
    # assert 0 < gaswater.crosspoint() < 1

    check_endpoints(gaswater, krwend, krwmax, krgend)
    ####################################
    # Do it over again, but with LET:
    gaswater.add_LET_gas(t=1.1, krgend=krgend)
    gaswater.add_LET_water(t=1.1, krwend=krwend, krwmax=krwmax)
    assert gaswater.selfcheck()
    check_table(gaswater.gasoil.table)
    check_table(gaswater.wateroil.table)
    # Check endpoints for oil curve:
    check_endpoints(gaswater, krwend, krwmax, krgend)
    check_linear_sections(gaswater.wateroil)
    check_linear_sections(gaswater.gasoil)
    # assert 0 < gaswater.crosspoint() < 1


def test_swfn():
    """Test that we can dump SWFN without giving gas relperm"""
    gaswater = GasWater(h=0.1)
    gaswater.add_corey_water()
    swfnstr = gaswater.SWFN()
    assert "SWFN" in swfnstr
    assert len(swfnstr) > 15
    assert "krw = krg" not in swfnstr


def test_sgfn():
    """Test that we can dump SGFN without giving water relperm"""
    gaswater = GasWater(h=0.1)
    gaswater.add_corey_gas()
    sgfnstr = gaswater.SGFN()
    assert "SGFN" in sgfnstr
    assert len(sgfnstr) > 15
    assert "krw = krg" not in sgfnstr


def test_linearsegments():
    """Made for testing the linear segments during
    the resolution of issue #163"""
    gaswater = GasWater(h=0.01, swl=0.1, swcr=0.3, sgrw=0.3)
    gaswater.add_corey_gas(ng=10, krgend=0.5)
    gaswater.add_corey_water(nw=10, krwend=0.5)
    check_table(gaswater.gasoil.table)
    check_table(gaswater.wateroil.table)
    check_linear_sections(gaswater.wateroil)
    check_linear_sections(gaswater.gasoil)


def test_gaswater_linear():
    """Test linear gaswater curves (linear as in giving only
    two saturation points to Eclipse)"""
    gaswater = GasWater(h=1)
    gaswater.add_corey_water()
    gaswater.add_corey_gas()
    swfnstr = gaswater.SWFN(header=False)
    check_table(gaswater.wateroil.table)
    check_table(gaswater.gasoil.table)
    check_linear_sections(gaswater.wateroil)
    check_linear_sections(gaswater.gasoil)
    assert isinstance(swfnstr, str)
    assert swfnstr
    assert len(gaswater.wateroil.table) == 2
    # assert np.isclose(gaswater.crosspoint(), 0.5)

    # What if there is no space for our choice of h?
    # We should be able to initialize nonetheless
    # (a warning could be given)
    gaswater = GasWater(swl=0.1, h=1)
    gaswater.add_corey_water()
    gaswater.add_corey_gas()
    check_table(gaswater.wateroil.table)
    check_table(gaswater.gasoil.table)
    check_linear_sections(gaswater.wateroil)
    check_linear_sections(gaswater.gasoil)
    assert len(gaswater.wateroil.table) == 2
    assert len(gaswater.gasoil.table) == 2
    assert np.isclose(gaswater.wateroil.table["SW"].min(), 0.1)
    assert np.isclose(gaswater.wateroil.table["SW"].max(), 1.0)
    # assert np.isclose(gaswater.crosspoint(), 0.55)


def test_crosspoint():
    """Test the crosspoint computation (on edge cases)"""
    gaswater = GasWater(swl=0.0, sgrw=0.0, sgcr=0.0, h=0.1)
    gaswater.add_corey_water(nw=1)
    gaswater.add_corey_gas(ng=1)
    assert np.isclose(gaswater.crosspoint(), 0.5)

    assert "-- krw = krg @ sw=0.5" in gaswater.SWFN()

    gaswater = GasWater(swl=0.5, sgrw=0.0, sgcr=0.0, h=0.1)
    gaswater.add_corey_water(nw=1)
    gaswater.add_corey_gas(ng=1)
    assert np.isclose(gaswater.crosspoint(), 0.75)

    gaswater = GasWater(swl=0.0, sgrw=0.5, sgcr=0.0, h=0.1)
    gaswater.add_corey_water(nw=1)
    gaswater.add_corey_gas(ng=1)
    assert np.isclose(gaswater.crosspoint(), 0.3333333)

    gaswater = GasWater(swl=0.0, sgrw=0.5, sgcr=0.5, h=0.1)
    gaswater.add_corey_water(nw=1)
    gaswater.add_corey_gas(ng=1)
    assert np.isclose(gaswater.crosspoint(), 0.25)

    gaswater = GasWater(swl=0.0, sgrw=0.5, sgcr=0.5, h=0.1)
    gaswater.add_corey_water(nw=1, krwend=0.5)
    gaswater.add_corey_gas(ng=1, krgend=0.5)
    assert np.isclose(gaswater.crosspoint(), 0.25)

    # Test warning/error situation:
    del gaswater.wateroil.table["KRW"]
    assert gaswater.crosspoint() is None


def test_gaswater_pc():
    """Test that capillary pressure can be added to GasWater.

    The GasWater object is calling up the code in WaterOil, which is
    tested more thorougly, in this test function we need to make
    sure the functionality is in place."""
    gaswater = GasWater(swl=0.1, h=0.2)
    gaswater.add_corey_water()
    gaswater.add_corey_gas()
    gaswater.add_simple_J()
    assert gaswater.wateroil.table["PC"].abs().sum() > 0
    swfn = gaswater.SWFN()
    assert "Simplified J-function" in swfn
    assert "0.1000000 0.0000000 0.23266" in swfn  # this is the first row.
    sat_table_str_ok(swfn)

    sgfn = gaswater.SGFN()
    # Capillary pressure in SGFN must always be zero for GasWater.
    assert "Zero capillary pressure" in sgfn
    sat_table_str_ok(sgfn)

    # Overwrite to zero:
    gaswater.add_simple_J(drho=0)
    swfn = gaswater.SWFN()
    assert "0.1000000 0.0000000 0.0000000" in swfn  # first row
    sat_table_str_ok(sgfn)

    # Petrophysical pc:
    gaswater.add_simple_J_petro(a=1, b=-1)
    swfn = gaswater.SWFN()
    assert "petrophysical version" in swfn
    assert "0.1000000 0.0000000 0.014715" in swfn  # first row


@pytest.mark.skipif(sys.platform == "win32", reason="Plotting not stable on windows")
def test_plotting(mocker):
    """Test that plotting code pass through (nothing displayed)"""
    mocker.patch("matplotlib.pyplot.show", return_value=None)
    gaswater = GasWater(swl=0.1, h=0.1)
    gaswater.add_corey_gas()
    gaswater.add_corey_water()
    gaswater.plotkrwkrg(mpl_ax=matplotlib.pyplot.subplots()[1])
    gaswater.plotkrwkrg(logyscale=True, mpl_ax=matplotlib.pyplot.subplots()[1])
    gaswater.plotkrwkrg(mpl_ax=None)


def test_comments():
    """Test that the outputters include endpoints in comments"""
    gaswater = GasWater(h=0.3)
    gaswater.add_corey_water()
    gaswater.add_corey_gas()
    swfn = gaswater.SWFN()
    assert "--" in swfn
    assert "pyscal: " in swfn  # part of version string
    assert "swirr=0" in swfn
    assert "swcr=0" in swfn
    assert "swl=0" in swfn
    assert "sgrw=0" in swfn
    assert "sgcr=0" in swfn
    assert "nw=2" in swfn
    assert "krwend=1" in swfn
    assert "Corey" in swfn
    assert "krw = krg @ sw=0.5" in swfn
    assert "Zero capillary pressure" in swfn
    assert "SW" in swfn
    assert "KRW" in swfn
    assert "KRGW" not in swfn
    assert "PC" in swfn

    sgfn = gaswater.SGFN()
    assert "--" in sgfn
    assert "pyscal: " in sgfn  # part of version string
    assert "swirr=0" in sgfn
    assert "sgcr=0" in sgfn
    assert "swl=0" in sgfn
    assert "ng=2" in sgfn
    assert "krgend=1" in sgfn
    assert "Corey" in sgfn
    assert "krw = krg @ sw=0.5" in sgfn
    assert "Zero capillary pressure" in sgfn
    assert "SG" in sgfn
    assert "KRW" not in sgfn
    assert "KRG" in sgfn
    assert "PC" in sgfn



"""Test module for the saturation ranges in WaterOil objects"""

import contextlib

import hypothesis.strategies as st
from hypothesis import given

from pyscal import WaterOil
from pyscal.constants import EPSILON, SWINTEGERS
from pyscal.utils.testing import check_table, float_df_checker


@given(
    st.floats(),
    st.floats(),
    st.floats(),
    st.floats(),
    st.floats(),
    st.floats(min_value=-0.1, max_value=2),
    st.text(),
)
def test_wateroil_random(swirr, swl, swcr, sorw, socr, h, tag):
    """Shoot wildly with arguments, the code should throw ValueError
    or AssertionError when input is invalid, but we don't want other crashes"""
    with contextlib.suppress(ValueError, AssertionError):
        WaterOil(swirr=swirr, swl=swl, swcr=swcr, sorw=sorw, socr=socr, h=h, tag=tag)


@given(
    st.floats(min_value=0, max_value=0.1),
    st.floats(min_value=0, max_value=0.15),
    st.floats(min_value=0, max_value=0.3),
    st.floats(min_value=0, max_value=0.3),
    st.floats(min_value=0, max_value=0.05),
    st.floats(min_value=0.01, max_value=0.2),
    st.text(),
)
def test_wateroil_normalization(swirr, swl, swcr, sorw, socr_add, h, tag):
    """Shoot with more realistic values and test normalized saturations"""
    wateroil = WaterOil(
        swirr=swirr, swl=swl, swcr=swcr, sorw=sorw, socr=sorw + socr_add, h=h, tag=tag
    )
    assert not wateroil.table.empty
    assert not wateroil.table.isna().to_numpy().any()

    # Check that son is 1 at swl:
    assert float_df_checker(wateroil.table, "SW", wateroil.swl, "SON", 1)

    # Check that son is 0 at socr:
    if wateroil.socr > h:
        assert float_df_checker(wateroil.table, "SW", 1 - wateroil.socr, "SON", 0)

    # Check that swn is 0 at swcr:
    assert float_df_checker(wateroil.table, "SW", wateroil.swcr, "SWN", 0)
    # Check that swn is 1 at 1 - sorw
    if wateroil.sorw > 1 / SWINTEGERS:
        assert float_df_checker(wateroil.table, "SW", 1 - wateroil.sorw, "SWN", 1)

    # Check that swnpc is 0 at swirr and 1 at 1:
    if wateroil.swirr >= wateroil.swl + h:
        assert float_df_checker(wateroil.table, "SW", wateroil.swirr, "SWNPC", 0)
    else:
        # Let this go, when swirr is too close to swl. We
        # are not guaranteed to have sw=swirr present
        pass

    assert float_df_checker(wateroil.table, "SW", 1.0, "SWNPC", 1)


@given(st.floats(min_value=0, max_value=1 - EPSILON))
def test_wateroil_swir(swirr):
    """Check that the saturation values are valid for all swirr"""
    wateroil = WaterOil(swirr=swirr)
    check_table(wateroil.table)


@given(st.floats(min_value=0, max_value=1 - EPSILON))
def test_wateroil_swl(swl):
    """Check that the saturation values are valid for all swl"""
    wateroil = WaterOil(swl=swl)
    check_table(wateroil.table)


@given(st.floats(min_value=0, max_value=1 - EPSILON))
def test_wateroil_swcr(swcr):
    """Check that the saturation values are valid for all swcr"""
    wateroil = WaterOil(swcr=swcr)
    check_table(wateroil.table)


@given(st.floats(min_value=0, max_value=1 - EPSILON))
def test_wateroil_sorw(sorw):
    """Check that the saturation values are valid for all sorw"""
    wateroil = WaterOil(sorw=sorw)
    check_table(wateroil.table)


@given(st.floats(min_value=0, max_value=1 - EPSILON))
def test_wateroil_socr(socr):
    """Check that the saturation values are valid for all socr"""
    wateroil = WaterOil(socr=socr)
    check_table(wateroil.table)


@given(st.floats(min_value=0, max_value=1), st.floats(min_value=0, max_value=1))
def test_wateroil_dual(param1, param2):
    """Test combination of 2 floats as parameters"""
    try:
        wateroil = WaterOil(swl=param1, sorw=param2)
        check_table(wateroil.table)
        # Will fail when swl > 1 - sorw
    except AssertionError:
        pass

    try:
        wateroil = WaterOil(swl=param1, swirr=param2)
        check_table(wateroil.table)
    except AssertionError:
        pass

    try:
        wateroil = WaterOil(swcr=param1, sorw=param2)
        check_table(wateroil.table)
    except AssertionError:
        pass

    try:
        wateroil = WaterOil(swirr=param1, sorw=param2)
        check_table(wateroil.table)
    except AssertionError:
        pass
